"""
Since:
    1.6.0
"""
import hashlib
import os
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

from box import Box
from tqdm import tqdm

from inspyre_toolbox.common.errors import InvalidParameterCombinationError, MissingRequiredParameterError
from inspyre_toolbox.common.types import File as FileType
from inspyre_toolbox.conversions.bytes import get_lowest_unit_size
from inspyre_toolbox.filesystem.errors import NeedsProcessingError
from inspyre_toolbox.filesystem.file import File
from inspyre_toolbox.filesystem.file import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.filesystem.file.helpers import get_path_list_from_list_of_file_objects
from inspyre_toolbox.humanize import Numerical
from inspyre_toolbox.log_engine import Loggable
from inspyre_toolbox.path_man import gather_files_in_dir, prepare_path, provision_path, provision_paths
from inspyre_toolbox.syntactic_sweets.classes.decorators.type_validation import validate_type
from inspyre_toolbox.syntactic_sweets.locks import flag_lock

MOD_LOGGER = PARENT_LOGGER.get_child('collection')


class NeedsReprocessingTag:

    def __init__(self, value: bool = False):
        self.__value = value if isinstance(value, bool) else None
        if self.__value is None:
            raise ValueError("Value must be a boolean.")

    def __get__(self, instance, owner):
        return self.__value

    def __set__(self, instance, value):
        if not isinstance(value, bool):
            raise ValueError("Value must be a boolean.")
        self.__value = value


@dataclass
class FileCollection(Loggable):
    """
    A class for managing a collection of files.

    Properties:
        paths (list):
            A list of file paths.

        total_size (int):
            The total size of the collection, in bytes.

        total_files (int):
            The total number of files in the collection.

        extensions (dict):
            A dictionary of extensions with classes for total size and number of files.

        needs_reprocessing (bool):
            A flag indicating whether the collection needs reprocessing.

    Methods:
        get_total_size_in_lowest_unit:
            Get the total size of the collection in the lowest unit.

        get_total_extension_size_in_lowest_unit:
            Get the total size of a specific extension in the lowest unit.

        remove_file:
            Remove a file from the collection.

        reprocess_files:
            Reprocess the files in the collection.

    Examples:
        >>> collection = FileCollection(paths=['/path/to/file1', '/path/to/file2'])
        >>> collection.total_size
        123456  # Total size of the collection, in bytes.
    """
    paths: list
    total_size: int = field(init=False, default=0)
    total_files: int = field(init=False, default=0)
    extensions: dict = field(init=False, default_factory=dict)

    def __init__(self, paths: List[Union[str, Path]] = None, auto_process: bool = False, do_not_use_progress_bar:
    bool = False):
        """
        Initialize the FileCollection with a list of file paths.

        `FileCollection` is a class for managing a collection of files. It takes a list of file paths and creates a class
        representing a collection of files with classes for total size and number of files and a dictionary of extensions
        with classes for total size and number of files.

        Parameters:
            paths (list):
                A list of file paths.
                
            auto_process (bool):
                A flag indicating whether to automatically process the files when the collection is created. If True, the
                files will be processed automatically. If False, the files will not be processed automatically. Default is
                False.
                
            do_not_use_progress_bar (bool):
                A flag indicating whether to disable the progress bar when processing the files. If True, the progress bar
                will be disabled. If False, the progress bar will be enabled. Default is False.

        Returns:
            None
        """
        super().__init__(MOD_LOGGER)
        self.__initialized = False
        self._files = {
                'local':  {
                        'files':      [],
                        'total_size': 0,
                        },
                'remote': {
                        'files':      [],
                        'total_size': 0,
                        }
                }
        self.__checksums = None
        self.__use_progress_bar = True
        self._files = Box(self._files)
        self._checksums_gathered = False
        self._getting_checksums = False
        self._needs_processing = False
        self._files_gathered = False
        self._processing = False
        self.__file_objects = {}
        self.__paths = paths
        self.__file_object_hash = None

        if self.paths:
            self._files_gathered = True
            self._needs_processing = True

        self.total_size = 0
        self.total_files = 0
        self.extensions = {}
        self.__needs_reprocessing = False

        if auto_process:
            self.process_files()

        self.__initialized = True

    @property
    def checksums(self):
        if self.__checksums is None:
            self.get_all_checksums()
        return self.__checksums

    @property
    def files(self):
        if self.needs_processing and not self._processing:
            raise NeedsProcessingError("Files need to be processed before they can be accessed.")

        return self._files

    @files.setter
    @validate_type(dict)
    def files(self, value: dict):
        if not self._processing:
            raise ValueError("Files cannot be set directly. Use the process_files method to set the files.")
        self._files = value

    @property
    def files_gathered(self) -> bool:
        """
        Returns whether the files have been gathered.

        This property returns a boolean value indicating whether the files have been gathered. If the files have been
        gathered, this property will return `True`. If the files have not been gathered, it will return `False`. The
        property is read-only.

        Returns:
            bool:
                True if the files have been gathered, False otherwise.
        """
        return self._files_gathered

    @property
    def file_names(self):
        return list(self.file_objects.keys())

    @property
    def file_objects(self):
        return self.__file_objects

    @property
    def file_objects_hash(self):
        if not self.__file_object_hash:
            self.__file_object_hash = self.get_file_object_hash()

        return self.__file_object_hash

    @property
    def needs_processing(self):
        return self._needs_processing

    @property
    def needs_reprocessing(self) -> bool:
        """
        Returns whether the collection needs reprocessing.

        This property returns a boolean value indicating whether the collection needs reprocessing. If the collection has
        been modified since the last processing, this property will return `True`. If the collection has not been modified,
        it will return `False`. The property is read-only.

        Returns:
            bool:
                True if the collection needs reprocessing, False otherwise.
        """
        return self.__needs_reprocessing

    @needs_reprocessing.setter
    def needs_reprocessing(self, value: bool):
        """
        Sets whether the collection needs reprocessing.

        This method sets the `needs_reprocessing` attribute of the collection. If the value is `True`, the collection will
        be reprocessed the next time it is accessed. If the value is `False`, the collection will not be reprocessed. The
        method is used to indicate that the collection has been modified and needs to be reprocessed.

        Parameters:
            value (bool):
                The value to set the `needs_reprocessing` attribute to. True if the collection needs reprocessing, False
                otherwise.

        Returns:
            None
        """
        self.__needs_reprocessing = value

    @property
    def paths(self):
        if not self.__paths and self.file_objects:
            self.__paths = get_path_list_from_list_of_file_objects(self.file_objects)

        return self.__paths

    @paths.setter
    def paths(self, new: List[Union[str, Path]]):
        self.__paths = new

    @property
    def path_strings(self):
        return [str(path) for path in self.paths]

    @property
    def total_local_size(self):
        return self.files['local']['total_size']

    @property
    def total_remote_size(self):
        return self.files['remote']['total_size']

    @property
    def use_progress_bar(self):
        return self.__use_progress_bar

    def backup_all_files(
            self,
            backup_dir: Optional[Union[str, Path]],
            archive: bool = False,
            delete_backups_on_archive: bool = False,
            delete_originals: bool = False,
            skip_archive_confirmation: bool = False,
            skip_backup_delete_confirmation: bool = False,
            skip_delete_originals_confirmation: bool = False,
            skip_all_confirmations: bool = False,
            **kwargs):
        """
        Backup all files in the collection.

        This method backs up all files in the collection to a backup directory. It includes the option to archive
        the backed-up files, delete the backups after archiving, and delete the original files after backing them up.
        Confirmation prompts are enabled by default for archiving, deletion of backups, and deletion of originals,
        but can be skipped individually or altogether.

        Parameters:
            backup_dir (Union[str, Path]):
                The backup directory. This can be a string or a Path object. If a string is provided, it will be converted
                to a Path object, expanded, resolved, and absolute.

            archive (bool):
                Whether the backup should be archived after the files are copied. Default is False.

            delete_backups_on_archive (bool):
                A flag indicating whether to delete the backup directory after archiving. Default is False.

            delete_originals (bool):
                Whether to delete the original files after backing them up. Default is False.

            skip_archive_confirmation (bool):
                Whether to skip the confirmation prompt before archiving. Default is False.

            skip_backup_delete_confirmation (bool):
                Whether to skip the confirmation prompt before deleting the backups. Default is False.

            skip_delete_originals_confirmation (bool):
                Whether to skip the confirmation prompt before deleting the original files. Default is False.

            skip_all_confirmations (bool):
                If True, all confirmation prompts will be skipped, overriding the other skip confirmation parameters.
                Default is False.

            **kwargs:
                Additional keyword arguments.

        Returns:
            Path: The path to the backup directory or archive.

        Raises:
            ValueError:
                If the backup directory cannot be created.

            FileNotFoundError:
                If a file in the collection does not exist.

            FileExistsError:
                If a file in the backup directory already exists.

            PermissionError:
                If the user does not have permission to write to the backup directory.

            OSError:
                If an error occurs while creating the backup directory or copying files.
        """

        # Convert backup_dir to Path object if it's a string
        backup_dir = Path(backup_dir).expanduser().resolve()

        # Create the backup directory if it doesn't exist
        try:
            backup_dir.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise ValueError(f"Cannot create backup directory: {backup_dir}") from e

        backup_files = []
        dest = Path(backup_dir)
        for file_path in self.files:
            try:
                # Copy each file to the backup directory
                dest = backup_dir / file_path.name
                shutil.copy2(file_path, dest)
                backup_files.append(dest)
            except FileNotFoundError:
                raise FileNotFoundError(f"File not found: {file_path}")
            except FileExistsError:
                raise FileExistsError(f"File already exists in backup: {dest}")
            except PermissionError:
                raise PermissionError(f"No permission to copy {file_path} to {dest}")
            except OSError as e:
                raise OSError(f"Error copying {file_path} to {dest}: {e}")

        # Archive the backup directory if requested
        if archive:
            if not skip_all_confirmations and not skip_archive_confirmation:
                confirm = input("Are you sure you want to archive the backup directory? (y/n): ")
                if confirm.lower() != 'y':
                    print("Archiving skipped.")
                    return backup_dir

            archive_path = shutil.make_archive(str(backup_dir), 'zip', root_dir=str(backup_dir))

            if delete_backups_on_archive:
                if not skip_all_confirmations and not skip_backup_delete_confirmation:
                    confirm = input(
                            f"Are you sure you want to delete the backup directory {backup_dir} after archiving? (y/n): ")
                    if confirm.lower() != 'y':
                        print("Backup directory not deleted.")
                        return Path(archive_path)
                shutil.rmtree(backup_dir)

            return Path(archive_path)

        # Delete the original files if requested
        if delete_originals:
            if not skip_all_confirmations and not skip_delete_originals_confirmation:
                confirm = input("Are you sure you want to delete the original files? (y/n): ")
                if confirm.lower() != 'y':
                    print("Original files not deleted.")
                    return backup_dir

            for file_path in self.files:
                try:
                    os.remove(file_path)
                except FileNotFoundError:
                    raise FileNotFoundError(f"Original file not found: {file_path}")
                except PermissionError:
                    raise PermissionError(f"No permission to delete original file: {file_path}")
                except OSError as e:
                    raise OSError(f"Error deleting original file {file_path}: {e}")

        return backup_dir

    def bucketize_file(self, file: FileType):
        """
        Bucketize a file.

        This method "bucketizes" a file into the local or remote bucket based on whether the file has
        a recall attribute. If the file has a recall attribute, it is bucketized into the remote
        bucket. If the file does not have a recall attribute, it is bucketized into the local bucket.

        Parameters:
            file

        Returns:

        """
        try:
            remote = file.has_recall_attribute
        except NotImplementedError:
            remote = False

        bucket = self.files['remote'] if remote else self.files['local']
        bucket['files'].append(file)
        bucket['total_size'] += file.size_in_bytes

    def get_file_object_hash(self):
        # Create a new hash object
        hash_obj = hashlib.sha256()

        if self.needs_processing:
            raise NeedsProcessingError("Files need to be processed before a hash can be generated.")

        for obj in self.file_objects.values():
            # Convert each object to a hashable representation (e.g., a string)
            obj_repr = repr(obj).encode('utf-8')

            # Update the hash object with this representation
            hash_obj.update(obj_repr)

        # Return the hexadecimal digest of the combined hash
        return hash_obj.hexdigest()


    def process_files(self):
        """
        Process the files in the collection.

        This method processes the files in the collection. It calculates the total size of the collection, the total number
        of files, and the total size of each extension in the collection. It populates the `total_size`, `total_files`, and
        `extensions` attributes of the class.

        Returns:
            None
        """
        from inspyre_toolbox.filesystem.file.helpers import get_file_object

        if all(isinstance(path, File) for path in self.paths):
            self.paths = [file.path for file in self.paths]
        elif all(isinstance(path, (str, Path)) for path in self.paths):
            self.paths = provision_paths(self.paths)

        with flag_lock(self, 'processing'):


            for path_str in self.paths:
                if isinstance(path_str, Path):
                    path_str = str(path_str)
                elif isinstance(path_str, File):
                    path_str = path_str.path

                file_name = Path(path_str).name

                file = get_file_object(path_str, skip_path_provision=True)

                self.file_objects[file_name] = file

                self.bucketize_file(file)

                if file.extension not in self.extensions:
                    self.extensions[file.extension] = {
                            'total_size':  0,
                            'total_files': 0,
                            }

                self.total_size += file.size_in_bytes
                self.total_files += 1

                self.extensions[file.extension]['total_size'] += file.size_in_bytes
                self.extensions[file.extension]['total_files'] += 1

        self._needs_processing = False
        self.get_file_object_hash()

    def find_file(
            self,
            path: Optional[Union[str, Path]],
            name: Optional[str] = None,
            include_remote: bool = False
            ) -> Union[File, None]:
        """
        Find a file in the collection.

        This method finds a file in the collection by its path. It takes a file path as an argument and returns the file
        object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            path (Optional[Union[str, Path]]):
                The file path to find.

            name (Optional[str]):
                The name of the file to find. If the name is provided, the method will search for the file by name instead

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files.

        Warning:
            = If the `name` and `path` parameters are both provided values, an `In


        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.
        """
        #     log = self.create_logger()

        if path and name:
            raise InvalidParameterCombinationError("Cannot provide both path and name parameters.")
        elif not path and not name:
            raise MissingRequiredParameterError("Either path or name parameter is required.")

        if name:
            return self.find_file_by_name(name, include_remote=include_remote, case_sensitive=False)
        elif path:
            return self.find_file_by_path(path, include_remote=include_remote)

        return None

    def find_file_by_name(self, name: str, include_remote: bool = False, case_sensitive: bool = False) -> Union[
        File, None]:
        # sourcery skip: swap-nested-ifs
        """
        Find a file in the collection by name.

        This method finds a file in the collection by its name. It takes a file name as an argument and returns the file
        object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            name (str):
                The name of the file to find.

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        if not case_sensitive:
            name = name.lower()

        if found_local := self.find_local_by_name(name, case_sensitive=case_sensitive):
            return found_local

        if include_remote:
            if found_remote := self.find_remote_by_name(name, case_sensitive=case_sensitive):
                return found_remote

        return None

    def find_local_by_name(self, name: str, case_sensitive: bool = False) -> Union[File, None]:
        """
        Find a local file in the collection by name.

        This method finds a local file in the collection by its name. It takes a file name as an argument and returns the
        file object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            name (str):
                The name of the file to find.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        for file in self.files.local.files:
            if case_sensitive:
                if file.name == name:
                    return file

            elif file.name.lower() == name:
                return file

    def find_remote_by_name(self, name: str, case_sensitive: bool = False) -> Union[File, None]:
        """
        Find a remote file in the collection by name.

        This method finds a remote file in the collection by its name. It takes a file name as an argument and returns the
        file object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            name (str):
                The name of the file to find.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        for file in self.files.remote.files:
            if case_sensitive:
                if file.name == name:
                    return file

            elif file.name.lower() == name:
                return file

    def find_file_by_path(self, path: Union[str, Path], include_remote: bool = False) -> Union[File, None]:
        """
        Find a file in the collection by path.

        This method finds a file in the collection by its path. It takes a file path as an argument and returns the file
        object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            path (Union[str, Path]):
                The file path to find.

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        path = provision_path(path)

        if found_local := self.find_local_by_path(path):
            return found_local

        if include_remote:
            if found_remote := self.find_remote_by_path(path):
                return found_remote

        return None

    def find_local_by_path(self, path: Union[str, Path]) -> Union[File, None]:
        """
        Find a local file in the collection by path.

        This method finds a local file in the collection by its path. It takes a file path as an argument and returns the
        file object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            path (Union[str, Path]):
                The file path to find.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        path = provision_path(path)

        return next(
                (file for file in self.files.local.files if file.path == path), None
                )

    def find_remote_by_path(self, path: Union[str, Path]) -> Union[File, None]:
        """
        Find a remote file in the collection by path.

        This method finds a remote file in the collection by its path. It takes a file path as an argument and returns the
        file object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            path (Union[str, Path]):
                The file path to find.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.

        """
        path = provision_path(path)

        return next(
                (file for file in self.files.remote.files if file.path == path), None
                )

    def get_all_checksums(self, with_progress_bar=use_progress_bar):
        if self.needs_processing:
            raise NeedsProcessingError("Files need to be processed before checksums can be accessed.")

        with flag_lock(self, 'getting_checksums'):
            self.__checksums = {}

            if with_progress_bar:
                paths = tqdm(self.file_objects.items(), desc="Calculating checksums", unit="file")
            else:
                paths = self.file_objects.items()

            for name, file in paths:
                self.__checksums[name] = file.get_checksum()

    def get_total_size_in_lowest_unit(self) -> tuple[Union[int, float], str]:
        """
        Get the total size of the collection in the lowest unit with a size greater than or equal to 1.

        This method calculates the total size of the collection in the lowest unit with a size greater than or equal
        to 1. It returns the size and the unit as a tuple. The unit is a string representing the unit of the size.

        Returns:
            tuple[Union[int, float], str]:
                The total size of the collection and the unit.

        """
        return get_lowest_unit_size(self.total_size)

    def get_total_extension_size_in_lowest_unit(
            self,
            extension: str,
            return_as_string: bool = False
            ) -> tuple[Union[int, float], str]:
        """
        Get the total size of a specific extension in the lowest unit with a size greater than or equal to 1.

        Parameters:
            extension (str):
                The extension to get the size of.

            return_as_string (bool):
                A flag indicating whether to return the size as a string. If True, the size will be returned as a string
                with the unit. If False, the size will be returned as a tuple with the size and the unit.

        Returns:
            tuple[Union[int, float], str]:
                The total size of the extension and the unit.

        Raises:
            ValueError:
                If the extension is not found in the collection.
        """

        if not extension.startswith('.'):
            extension = f".{extension}"

        extension = extension.lower()

        if extension not in self.extensions:
            raise ValueError(f"Invalid extension: {extension}!")

        lowest_unit_size = get_lowest_unit_size(self.extensions[extension]['total_size'])

        if return_as_string:
            return Numerical(lowest_unit_size[0], noun=lowest_unit_size[1]).count_noun()
        return lowest_unit_size

    def remove_file(
            self,
            path: Optional[Union[Path, str]] = None,
            name: Optional[str] = None,
            **kwargs
            ):
        """
        Remove a file from the collection.

        This method removes a file from the collection. It takes a file path as an argument and removes the file from
        the collection. It then reprocesses the files in the collection to update the total size, total number of files,
        and total size of each extension.

        Parameters:
            path (Union[Path, str]):
                The file path to remove. If the path is provided, the method will search for the file by path. If the path
                is not provided, the method will search for the file by name.

            name (str):
                The name of the file to remove. If the name is provided, the method will search for the file by name instead

            **kwargs:
                Additional keyword arguments

        Note:
            This method does not delete the file from the file system. It only removes the file from the collection,
            and only if the file is in the collection. If the file is not in the collection, this method will raise a
            `KeyError`. After removing the file, the collection will be reprocessed to update the total size, total
            number of files, and total size of each extension.

        Note:
          - If the `name` and `path` parameters are both provided values, an `InvalidParameterCombinationError`
            will be raised.

          - If neither the `name` nor the `path` parameter is provided, a `MissingRequiredParameterError` will be
            raised.


        Raises:
            KeyError:
                If the file is not in the collection.

            InvalidParameterCombinationError:
                If the `name` and `path` parameters are both provided values.

            MissingRequiredParameterError:
                If neither the `name` nor the `path` parameter is provided.

        Returns:
            None
        """
        # log = self.create_child_logger()

        if path and name:
            raise InvalidParameterCombinationError("Cannot provide both path and name parameters.")
        elif not path and not name:

            raise MissingRequiredParameterError("Either path or name parameter is required.")

        old_paths = self.paths.copy()

        if name:
            self.remove_file_by_name(name, **kwargs)
        elif path:
            self.remove_file_by_path(path, **kwargs)

        if old_paths != self.paths:
            print('Paths have changed.')
            self.__needs_reprocessing = True
            self.reprocess_files()

    def remove_file_by_name(
            self,
            name: str,
            include_remote: bool = False,
            remove_all: bool = False,
            case_sensitive: bool = False
            ):
        """
        Remove a file from the collection by name.

        This method removes a file from the collection by name. It takes a file name as an argument and removes the file
        from the collection. It then reprocesses the files in the collection to update the total size, total number of
        files, and total size of each extension.

        If you do not want to include remote files in the search, set the `include_remote` parameter to `False`. If you
        want to include remote files in the search, set the `include_remote` parameter to `True`. By default, remote files
        are not included in the search.

        Parameters:
            name (str):
                The name of the file to remove.

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files. Default is False.

            remove_all (bool):
                A flag indicating whether to remove all files with the specified name. If True, all files with the specified
                name will be removed. If False, only the first file with the specified name will be removed. Default is False.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

        Returns:
            None
        """
        if not case_sensitive:
            name = name.lower()

        self.remove_local_by_name(name, case_sensitive=case_sensitive, remove_all=remove_all)

        if include_remote:
            self.remove_remote_by_name(name, case_sensitive=case_sensitive, remove_all=remove_all)

    def remove_local_by_name(self, name: str, case_sensitive=False, remove_all: bool = False):
        """
        Remove a local file from the collection by name.

        This method removes a local file from the collection by name. It takes a file name as an argument and removes the
        file from the collection. It then reprocesses the files in the collection to update the total size, total number
        of files, and total size of each extension.

        Parameters:
            name (str):
                The name of the file to remove.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

            remove_all (bool):
                A flag indicating whether to remove all files with the specified name. If True, all files with the specified
                name will be removed. If False, only the first file with the specified name will be removed. Default is False.

        Returns:
            None
        """
        if not case_sensitive:
            name = name.lower()

        files = self.files.local.files
        paths = self.paths

        for file in files:
            if file.name == name:
                fp = file.path
                if file in files:
                    files.remove(file)
                if fp in paths:
                    paths.remove(fp)

                if name in self.file_objects:
                    del self.file_objects[name]

    def remove_remote_by_name(self, name: str, case_sensitive=False, remove_all: bool = False):
        """
        Remove a remote file from the collection by name.

        This method removes a remote file from the collection by name. It takes a file name as an argument and removes the
        file from the collection. It then reprocesses the files in the collection to update the total size, total number
        of files, and total size of each extension.

        Parameters:
            name (str):
                The name of the file to remove.

            case_sensitive (bool):
                A flag indicating whether the search should be case-sensitive. If True, the search will be case-sensitive.
                If False, the search will be case-insensitive. Default is False.

            remove_all (bool):
                A flag indicating whether to remove all files with the specified name. If True, all files with the specified
                name will be removed. If False, only the first file with the specified name will be removed. Default is False.

        Returns:
            None
        """
        if not case_sensitive:
            name = name.lower()

        if remove_all:
            files = [file for file in self.files.remote.files if file.name == name]
        else:
            files = [next((file for file in self.files.remote.files if file.name == name), None)]

        for file in files:
            self.remove_file(file.path)

    def remove_file_by_path(self, path: Union[str, Path], include_remote: bool = False):
        """
        Remove a file from the collection by path.

        This method removes a file from the collection by its path. It takes a file path as an argument and removes the
        file from the collection. It then reprocesses the files in the collection to update the total size, total number
        of files, and total size of each extension.

        Parameters:
            path (Union[str, Path]):
                The file path to remove.

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files. Default is False.

        Returns:
            None
        """
        target_path = provision_path(path)
        local_files = self.path_strings

        if str(target_path) in local_files:
            self.remove_local_by_path(target_path)

        if include_remote:
            self.remove_remote_by_path(target_path, do_not_provision=True)

    def remove_local_by_path(self, path: Union[str, Path], do_not_provision=False):
        """
        Remove a local file from the collection by path.

        This method removes a local file from the collection by its path. It takes a file path as an argument and removes
        the file from the collection. It then reprocesses the files in the collection to update the total size, total number
        of files, and total size of each extension.

        Parameters:
            path (Union[str, Path]):
                The file path to remove.

            do_not_provision (bool):
                A flag indicating whether to skip path provisioning.
                  - If True; the path will not be provisioned before removing the file.
                  - If False; the path will be provisioned before removing the file.
                  - Default is False

        Raises:
            FileNotFoundError:
                If the file is not found in the collection.

        Returns:
            None
        """
        if not do_not_provision:
            target_path = provision_path(path)

        target_path = str(target_path)
        local_files = self.files.local.files
        paths = self.paths

        for file in self.files.local.files:
            if str(file.path) == target_path:
                local_files.remove(file)
                paths.remove(file.path)
                return

    def remove_remote_by_path(self, path: Union[str, Path], do_not_provision=False):
        """
        Remove a remote file from the collection by path.

        This method removes a remote file from the collection by its path. It takes a file path as an argument and removes
        the file from the collection. It then reprocesses the files in the collection to update the total size, total number
        of files, and total size of each extension.

        Parameters:
            path (Union[str, Path]):
                The file path to remove.

        Returns:
            None
        """
        if not do_not_provision:
            target_path = provision_path(path)

        target_path = str(target_path)

        for file in self.files.remote.files:
            if file.path == target_path:
                self.remove_file(file.path)
                return

    def reprocess_files(self):
        """
        Reprocess the files in the collection.

        This method reprocesses the files in the collection. It recalculates the total size of the collection, the total
        number of files, and the total size of each extension in the collection. It populates the `total_size`, `total_files`,
        and `extensions` attributes of the class. This method is used to update the collection after files have been added
        or removed.

        Note:
            This method should be called after files have been added or removed from the collection. It will only reprocess
            the files if the `needs_reprocessing` attribute is set to `True`. After reprocessing, the `needs_reprocessing`
            attribute will be set to `False`.

        Returns:
            None

        """
        if self.needs_reprocessing:
            self.total_size = 0
            self.total_files = 0
            self.extensions = {}
            self.process_files()



    def __getitem__(self, key: Union[int, str]) -> Union[str, Dict[str, int]]:
        if isinstance(key, int):
            return str(self.paths[key])
        elif isinstance(key, str):
            if not key.startswith('.'):
                key = f".{key}"
            key = key.lower()
            if key not in self.extensions:
                raise KeyError(f"Extension {key} not found in file collection.")
            return self.extensions[key]
        else:
            raise TypeError("Key must be an integer (for path index) or a string (for extension).")

    def __str__(self):
        size, unit = self.get_total_size_in_lowest_unit()
        size_str = Numerical(size, noun=unit).count_noun()
        return f"Total size: {size_str}, Total files: {self.total_files}"


def create_file_collection(paths: list[Path], **kwargs) -> FileCollection:
    """
    Create a file collection.

    This function creates a file collection from a list of file paths. It takes a list of file paths as an argument and
    returns a :class:`FileCollection` object representing the collection of files.

    Parameters:
        paths (list[Path]):
            A list of file paths.

    Returns:
        FileCollection:
            A `FileCollection` object representing the collection of files.

    Since:
        v1.0.0
    """
    return FileCollection(paths, **kwargs)


def collect_files(root_dir: Union[str, Path], recursive=True, **kwargs):
    """
    Collect files in a directory into a :class:`FileCollection`.

    Parameters:
        root_dir (Union[str, Path]):
            The root directory from which to collect files.

        recursive (bool):
            A flag indicating whether to collect files recursively.

        **kwargs:
            Additional keyword arguments.

    Returns:
        FileCollection:
            A `FileCollection` object representing the collection of files.

    """
    root_dir = prepare_path(root_dir, do_not_create=True)
    files = gather_files_in_dir(root_dir, file_types=kwargs.pop('extensions'), recursive=recursive)
    return create_file_collection(
            files,
            **kwargs
            )



from typing import NewType


def __split_evenly(target):
    """

    Split a target list down the center.

    Arguments:
        target (list):
            The list you'd like to split.

    Returns:
        (list, list):
            Two lists that should be perfectly (or as perfectly as possible
            while keeping whole numbers) split in two while maintaining
            its order.

    """
    length = len(target)
    mi = length // 2

    list1 = target[:mi]
    list2 = target[mi:]

    return list1, list2


def __split_alt(target):
    """

    Split a given list into two columns in an alternating format.


    ========= ===========
      Col 1      Col 2
    ========= ===========
        1          2
    --------- -----------
        3          4
    ========= ===========

    Args:
        target:

    Returns:

    """
    list1 = []
    list2 = []

    for item in target:
        if target.index(item) % 2 == 0:
            list1.append(item)
        else:
            list2.append(item)

    return list1, list2


def split_list(target, split_method='middle'):
    """

    Split a given list using an indicted method.

    Args:
        target:
        split_method:

    Returns:

    """
    valid_split_methods = [
            'alternating_columns',
            'middle',
            ]
    """
    valid_split_methods:

        * alternating_columns
        * middle

    Note:
        The values that will be accepted for the 'split_method'
        parameter are as follows;



       * alternating_columns:

          Sort into a number of columns where ordering goes from
          left-to-right.

        * middle:

          Split a list in half results in two smaller lists containing
          elements from the original in the same order. (The Default
          parameter value)

    """
    if split_method.lower() not in valid_split_methods:
        raise ValueError(f"The value for 'split_method' must be one of; {', '.join(valid_split_methods)}")
    if split_method.lower() == 'middle':
        return __split_evenly(target)
    elif split_method.lower() == 'alternating_columns':
        return __split_alt(target)


ChunkifiedList = NewType('ChunkifiedList', list)


def chunkify(target: list, num_per: int) -> ChunkifiedList:
    ret_lst = lambda target, num_per: [target[i:i + num_per] for i in range(0, len(target), num_per)]

    return ChunkifiedList(ret_lst(target, num_per))



from datetime import datetime
from typing import Optional

from inspyre_toolbox.syntactic_sweets.classes.decorators.freeze import freeze_property
from inspyre_toolbox.syntactic_sweets.classes.decorators.type_validation import validate_type


class OptionLedgerEntry:
    """
    A class to represent a ledger entry for tracking changes in option values.

    Attributes:
    -----------
    option_name (str):
        The name of the option being changed. Required, cannot be None.

    old_value (Optional[Any]):
        The old value of the option before the change. Can be None.

    new_value (Optional[Any]):
        The new value of the option after the change. Cannot be None.

    timestamp (datetime):
        The time when the change occurred.
    """

    def __init__(self, option_name: str, old_value: Optional[object], new_value: Optional[object]):
        self.__initialized = False
        self.__option_name = None
        self.__old_value = None
        self.__new_value = None
        self.__timestamp = None

        self.option_name = option_name
        self.old_value = old_value
        self.new_value = new_value

        self.__timestamp = datetime.now()
        self.__initialized = True

    @property
    def initialized(self):
        return self.__initialized

    @property
    def new_value(self):
        return self.__new_value

    @new_value.setter
    @validate_type(object)
    def new_value(self, new_value):
        if self.initialized:
            raise AttributeError("Cannot modify 'new_value' after the entry has been initialized.")

        self.__new_value = new_value

    @property
    def old_value(self):
        return self.__old_value

    @old_value.setter
    @validate_type(object)
    def old_value(self, old_value):
        if self.initialized:
            raise AttributeError("Cannot modify 'old_value' after the entry has been initialized.")
        self.__old_value = old_value

    @property
    def option_name(self):
        return self.__option_name

    @option_name.setter
    @validate_type(str)
    @freeze_property
    def option_name(self, option_name):
        if self.initialized:
            raise AttributeError("Cannot modify 'option_name' after the entry has been initialized.")

        self.__option_name = option_name

    @property
    def timestamp(self):
        return self.__timestamp

    def __repr__(self):
        return (f"Option '{self.option_name}' changed from '{self.old_value}' "
                f"to '{self.new_value}' at {self.timestamp}")


class OptionLedger:
    """
    A class to track changes to option values.

    Attributes:
    -----------
    entries : List[OptionLedgerEntry]
        A list of all changes to option values.
    """

    def __init__(self):
        self.entries = []

    def add_entry(self, entry: OptionLedgerEntry):
        """Add a new entry to the ledger."""
        self.entries.append(entry)

    def get_entries(self):
        """Return all ledger entries."""
        return self.entries

    def __repr__(self):
        return "\n".join([str(entry) for entry in self.entries])



"""


Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/type_parser/boolean.py
 

Description:
    

"""
from typing import Any, Optional

DEFAULT_TRUTHY_VALUES = ['true', 't', 'yes', 'y', '1', 1, True]
DEFAULT_FALSEY_VALUES = ['false', 'f', 'no', 'n', '0', 0, False]
DEFAULT_BOOLEAN_TYPES = (int, str, bool)


def __prepare_boolean_values(true_false: bool, additional: list[Any]=None, exclude: list[Any]=None):

    # If true_false is True, we are preparing truthy values
    if true_false:
        values = DEFAULT_TRUTHY_VALUES.copy()
    else:
        # If true_false is False, we are preparing falsey values
        values = DEFAULT_FALSEY_VALUES.copy()

    # If additional is None, set it to an empty list
    if additional is None:
        additional = []
    else:
        # If additional is not a list, convert it to a list
        if not isinstance(additional, list):
            additional = list(additional)

    # If exclude is None, set it to an empty list
    if exclude is None:
        exclude = []
    else:
        # If exclude is not a list, convert it to a list
        if not isinstance(exclude, list):
            exclude = list(exclude)

    # Remove excluded values
    for value in exclude:
        if value in values:
            values.remove(value)
        elif isinstance(value, str):
            if value.lower() in values:
                values.remove(value.lower())

    for value in values:
        if isinstance(value, str):
            values.append(value.lower())

    value_types = [type(value) for value in values]

    return values, value_types



def parse_boolean(
        value,
        false_on_not_found=False,
        false_on_error=False,
        truthy_additional: Optional[list[Any]]=None,
        truthy_exclude: Optional[list] = None,
        falsey_additional: Optional[list[Any]]=None,
        falsey_exclude: Optional[list] = None
):
    truthy_values, truthy_types = __prepare_boolean_values(True, truthy_additional, truthy_exclude)
    falsey_values, falsey_types = __prepare_boolean_values(False, falsey_additional, falsey_exclude)



import ctypes
import sys


def run_as_admin():
    """
    Run the current script as an admin.

    Raises:
        PermissionError:
            If the current user is not an admin.

    Since:
        v1.6.0

    Example:
        >>> from inspyre_toolbox.sys_man.operating_system import run_as_admin
        >>> run_as_admin()
    """
    params = ' '.join(f'"{arg}"' for arg in sys.argv)
    ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, params, None, 1)


__all__ = ['run_as_admin']



"""
A module containing the StringInputOption class.

Classes:
    StringInputOption:
        A class to represent a string input option with a description.

Since:
    1.6.0
"""
from typing import Dict

from inspyre_toolbox.config_man.options import Option


class StringInputOption(Option):
    """
    A class to represent a string input option with a description.

    Attributes:
    -----------
    name (str):
        The name of the option.

    description (str):
        A description of what the option does.

    value (str):
        The current value of the option.

    """

    def set_value(self, value: str):
        """
        Set the value of the option, enforcing type checking.

        Arguments:
            value (str):
                The new value for the option.

        Returns:
            None

        Raises:
            TypeError:
                If the value is not a string.
        """
        if not isinstance(value, str):
            raise TypeError("StringInputOption requires a string value.")
        old_value = self.value
        self.value = value
        self.log_change(old_value, value)

    def to_dict(self) -> Dict:
        """
        Convert the StringInputOption to a dictionary.

        Returns:
            Dict:
                A dictionary representation of the StringInputOption. This dictionary can be used to recreate the
                option, and should contain the following keys:
                    - 'type': The type of the option ('StringInputOption').
                    - 'name': The name of the option.
                    - 'description': A description of what the option does.
                    - 'value': The current value of the option.

        """
        return {
                'type':        'StringInputOption',
                'name':        self.name,
                'description': self.description,
                'value':       self.value
                }

    @staticmethod
    def from_dict(data: Dict):
        """
        Create a StringInputOption from a dictionary. The dictionary should contain the following
        keys:
            - 'name': The name of the option.
            - 'description': A description of what the option does.
            - 'value': The current value of the option.

        Arguments:
            data (Dict):
                A dictionary containing the data needed to create the StringInputOption object. This dictionary should
                contain the keys listed above.

        Returns:
            StringInputOption:
                A new StringInputOption object created from the data in the dictionary.
        """
        return StringInputOption(
                name=data['name'],
                description=data['description'],
                value=data['value']
                )



"""
This module is a collection of tools for managing the system. It is designed to be cross-platform, but is currently only
implemented for Windows and Linux. The tools in this module are designed to help manage the system, such as checking if
the current user has admin privileges, or checking if the current user has access to the GUI.

Subpackages:
    operating_system:
        This subpackage contains functions for managing the operating system. It is designed to be cross-platform, but is
        currently only implemented for Windows and Linux.
"""
#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
from inspyre_toolbox.log_engine import ROOT_LOGGER as PARENT_LOGGER

MOD_LOGGER = PARENT_LOGGER.get_child('sys-man')

from .operating_system.checks import is_windows

if is_windows():
    from .operating_system.win32 import *
else:
    from .operating_system.linux import *

__all__ = [
        'IS_ADMIN',
        ]



"""
This module contains decorators for class methods that allow you to add aliases for them.

Decorators:
    method_alias:
        Decorator to add alias names to a decorated method.

    add_aliases:
        Decorator to add aliases to a class's methods.
"""


def method_alias(*alias_names):
    """
    A decorator that allows you to add aliases to a class's method.

    This decorator adds an attribute '_alias_names' to the method, which stores the alias names.
    When the 'add_aliases' decorator is applied to the class, it uses these alias names to create
    additional references to the method.

    Parameters:
        *alias_names (str, list[str]):
            The name(s) of the alias(es) to add. You can pass multiple alias names as arguments.

    Returns:
        method:
            The decorated method with an '_alias_names' attribute.

    Example Usage:
        @method_alias('alias1', 'alias2')
        def my_method(self):
            print("Method called")

    Since:
        v1.6.0
    """

    def method_decorator(meth):
        meth._alias_names = alias_names  # Add alias names as an attribute to the method
        return meth

    return method_decorator


def add_aliases(cls):
    """
    A decorator to add aliases to a class's methods.

    This decorator iterates over the class's methods and checks if they have an '_alias_names'
    attribute. For each alias name found in '_alias_names', it creates an alias method in the class.

    Parameters:
        cls (class):
            The class to add aliases to.

    Returns:
        class:
            The decorated class with alias methods added.

    Example Usage:
        @add_aliases
        class MyClass:

            @method_alias('alias1', 'alias2')
            def original_method(self):
                print("Original method called")

        # After decoration, MyClass will have 'alias1' and 'alias2' as aliases for 'original_method'.
        obj = MyClass()
        obj.original_method()  # Output: Original method called
        obj.alias1()           # Output: Original method called
        obj.alias2()           # Output: Original method called

    Since:
        v1.6.0
    """

    for name, method in list(cls.__dict__.items()):  # Iterate over class's attributes
        if hasattr(method, '_alias_names'):  # Check if method has _alias_names attribute
            for alias_name in method._alias_names:  # For each alias name
                setattr(cls, alias_name, method)  # Add the alias as a method of the class
    return cls



from typing import Dict, Optional

from inspyre_toolbox.config_man.options import Option


class YesNoOption(Option):

    def __init__(self, name: str, description: str, value: Optional[bool] = None, frozen: bool = False):
        super().__init__(name, description, value)
        self.frozen = frozen

    def set_value(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("YesNoOption requires a boolean value.")
        if self.frozen:
            raise ValueError(f"Option '{self.name}' is frozen and cannot be modified.")
        old_value = self.value
        self.value = value
        self.log_change(old_value, value)

    def freeze(self):
        self.frozen = True

    def unfreeze(self):
        self.frozen = False

    def to_dict(self) -> Dict:
        return {
                'type':        'YesNoOption',
                'name':        self.name,
                'description': self.description,
                'value':       self.value,
                'frozen':      self.frozen
                }

    @staticmethod
    def from_dict(data: Dict):
        return YesNoOption(
                name=data['name'],
                description=data['description'],
                value=data['value'],
                frozen=data['frozen']
                )



from pathlib import Path
from typing import Union

from inspyre_toolbox.filesystem.file import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.path_man import provision_path
from inspyre_toolbox.sys_man.operating_system.checks import is_windows

MOD_LOGGER = PARENT_LOGGER.get_child('attributes')

try:
    if not is_windows():
        raise NotImplementedError('The attributes module is only available on Windows.')

    RECALL_ON_DATA_ACCESS_ATTR = 0x00400000

    __all__ = ['file_has_recall_attribute']

except NotImplementedError as e:
    MOD_LOGGER.error(f'Error: {e}')
    raise e


def file_has_recall_attribute(
        file_path: Union[str, Path],
        do_not_provision: bool = False
        ) -> bool:
    """
    Check if a file has the 'recall on data access' attribute set.

    This function checks if a file has the 'recall on data access' attribute set. This attribute is
    used by Windows to mark files that should be recalled from a storage device when they are
    accessed. OneDrive uses this attribute to mark files that are stored in the cloud and not stored
    locally. This function will return True if the file has the attribute set, and False if it does
    not. This can be useful for finding out which files must be downloaded from the cloud before they
    can have operations performed on them.

    Note:
        This function is only available on Windows.

    Parameters:
        file_path (Union[str, Path]):
            The path to the file to check.

        do_not_provision (bool):
            Whether to provision the path before checking it.

    Returns:
        bool:
            True if the file has the 'recall on data access' attribute set, False otherwise.
    """
    _name = 'file_has_recall_attribute'

    def wrapped(file_path, do_not_provision):
        # Set up the logger for this function if it doesn't already exist, otherwise get the existing
        # logger from the parent logger.
        if not MOD_LOGGER.find_child_by_name(_name):
            log = MOD_LOGGER.get_child(_name)
        else:
            log = MOD_LOGGER.find_child_by_name(_name)[0]

        from win32api import GetFileAttributes

        if not do_not_provision:
            log.debug(f'Provisioning path: {file_path}')
            file_path = provision_path(file_path)
        else:
            log.debug(f'Not provisioning path: {file_path}')

        try:
            log.debug(f'Checking file attributes for {file_path}')
            attrs = GetFileAttributes(str(file_path))
            log.debug(f'File attributes for {file_path}: {attrs}')
            return attrs & RECALL_ON_DATA_ACCESS_ATTR
        except Exception as e:
            log.error(f'Error checking file attributes for {file_path}: {e}')
            return False

    return bool(wrapped(file_path, do_not_provision))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  9 01:23:36 2021

@author: taylor
"""

import typing
from decimal import Decimal
from typing import Callable, ContextManager, Optional, Union

from inflect import engine

from inspyre_toolbox.core_helpers.logging import ROOT_ISL_DEVICE
from inspyre_toolbox.humanize.errors import HumanizeErrors

# Instantiate inflect.engine as 'INF'
INF = engine()

NUM_ERR = HumanizeErrors.NumericalErrors


class NumericalStrings(object):
    """

    An object containing all the basic parts of numbers in english, written-out as strings.

    """
    less_than_20 = [
            "",
            "One",
            "Two",
            "Three",
            "Four",
            "Five",
            "Six",
            "Seven",
            "Eight",
            "Nine",
            "Ten",
            "Eleven",
            "Twelve",
            "Thirteen",
            "Fourteen",
            "Fifteen",
            "Sixteen",
            "Seventeen",
            "Eighteen",
            "Nineteen",
    ]
    """ A list of all the numbers  'in string form' from 'One' to 'Twenty' """

    tens = [
            "",
            "Ten",
            "Twenty",
            "Thirty",
            "Forty",
            "Fifty",
            "Sixty",
            "Seventy",
            "Eighty",
            "Ninety",
    ]
    """ A list of all the double-digit numbers ending in '0' written-out in string form """

    thousands = [
            "",
            "Thousand",
            "Million",
            "Billion"
    ]
    """ Contains thousand, million, and billion strings """


class Numerical(object):
    """

    A class that allows you to manipulate numbers with greater-flexibility.

    Note:
        When performing mathmatical

    """

    def __init__(self, number, noun: typing.Optional[str] = None, store_as_float=False):
        """

        Create a new instance, with a new number to manipulate

        Args:
            number (int|float): The subject number.
            noun: The noun to count.
        """
        # First, check whether 'number' is an integer or a float or not. If not; we raise an error

        self.__store_as_float = store_as_float

        if not isinstance(number, float) and store_as_float:
            try:
                number = int(number)
            except ValueError as e:
                raise NUM_ERR.NotANumberError(
                        var=number,
                        var_name="number",
                        msg=f"Type {type(number)} is not valid. Must be {type(1)} or {type(1.1)}",

                ) from e

            if store_as_float:
                number = float(number)

        self.__number = number

        self.__noun = noun

        self.__commified = self.commify()

        self.log_name = 'Inspyre-Toolbox.humanize.Numerical'

        self.cls_logger = ROOT_ISL_DEVICE.get_child(self.log_name)
        self.cls_logger.debug(f"Started logger: {self.log_name}")

    def count_noun(
            self,
            number: Optional[Union[float, str, int]] = None,
            noun: Optional[Union[str, None]] = None,
            save_number: Optional[bool] = True,
            save_noun: Optional[bool] = True,
            **kwargs: object
    ):
        """
        
        Args:
            
            number: 
            noun (Optional, str):
                The thing you want to describe the number of. Defaults to 'self.__noun'.
            
            save_number (Optional, bool):
                If you provided a value for the 'number' parameter this parameter's value will indicate whether the
                value of 'number' should replace 'self.number'. Defaults to bool(True).
            
            save_noun (Optional, bool):
                If you provided a value for the 'noun' parameter and self.__noun isn't NoneType and differs from the
                provided value 'save_noun' will indicate whether the new value for 'noun' should replace
                'self.__noun'. Defaults to bool(True).
            
            **kwargs:
            
                full_stop (bool):
                 Include a full-stop punctuation mark at the end of the final statement. Users that
                 are from America may know this punctuation mark better as 'period' (or; '.'. See
                 below for the 'period' parameter alias)
                 (Defaults to boo  l(False)

                period (bool):
                    Include a period mark at the end of the final statement. Users from some areas may
                    know this mark better as 'full-stop'
    
                noun (str):
                    Returns a string with the appropriate statement iy,.hy,i for annotating counts of various nouns.
    
                count (int/float, optional):
                    The number of 'noun' that exists.
                    (Defaults to 'Numerical.number')
    
                only_noun (bool, optional):
                    Return only the properly pluralized noun.
                    (Defaults to False)
    
                skip_commify (bool, optional):
                    (Ignored if either 'only_noun' or 'count_to_words' are True) Return the number as
                    part of the statement with commas added to the proper places.
                    (Defaults to False)
    
                to_words (bool, optional):
                    (Ignored if 'only_noun' is True) Modify the first part of the statement
                    (which contains the count) to words instead of numbers.
                    (Defaults to False.)
    
                capitalize (bool, optional):
                    Return the words for the count with the first letter capitalized.
                    (Defaults to False, ignored if 'count_to_words' is False.)
    
                round_num (int, optional):
                    Round the number returned to a given precision in decimal digits.
                    (Defaults to None(NoneType), ignored if 'only_noun' is True.)

        Returns:

        """

        number_changed = False
        noun_changed = False

        number_previous = None
        noun_previous = None

        if number is not None:
            try:
                float(number)
                if number != self.__number and not save_number:
                    number_previous = self.__number
                    number_changed = True

                self.__number = number
            except ValueError as e:

                warning = "If passing a noun alone to 'count_noun' you need to include the parameter name\n" \
                          "For example:\n" \
                          "\n" \
                          "number.count_noun(noun='banana')\n" \
                          "NOT\n" \
                          "number.count_noun('banana')"

                print(warning)

                raise ValueError(warning) from e

        if noun is not None:

            noun = str(noun)
            try:
                if self.__noun is not None and noun != self.__noun:
                    if not save_noun:
                        noun_changed = True
                        noun_previous = self.__noun

                    self.__noun = noun
            except ValueError as e:
                raise ValueError(f"'noun' should be {type(str())} not {type(noun)}") from e

        ret = self.__count_noun(noun=self.__noun, count=self.__number, **kwargs)

        if number_changed:
            self.__number = number_previous
            number_previous = None
            number_changed = False
        if noun_changed:
            self.__noun = noun_previous
            noun_previous = None
            noun_changed = None

        return ret

    @property
    def noun(self):
        """

        The value of 'Numerical.noun'.

        Returns:
            Numerical.noun (str):
                The value of 'Numerical.noun'
        """
        return self.__noun

    @noun.setter
    def noun(self, new_value):
        if isinstance(new_value, str):
            self.__noun = new_value
        else:
            raise ValueError("Can only set a string as 'noun'")

    @noun.deleter
    def noun(self):
        self.__noun = None

    @property
    def dec_num(self):
        return Decimal(self.number)

    @property
    def commified(self):
        return self.commify()

    @property
    def number(self):
        return self.__number

    @number.setter
    def number(self, new_value):
        log_name = f'{self.log_name}.number.Setter'
        log = ROOT_ISL_DEVICE.get_child(log_name)
        try:
            float(new_value)
            int(new_value)
        except ValueError as e:
            log.error(f"The new value must be of type 'int' or 'float' not {type(new_value)}")

        self.__number = new_value
        log.info("New value set for 'Numerical.number'")
        self.commify(int(new_value))

    @property
    def store_as_float(self):
        return self.__store_as_float

    @store_as_float.setter
    def store_as_float(self, opt):
        """

        Set the 'store_as_float' property.

        Numerical.store_as_float is a settings property that dictates if the principle number of the Numerical class
        after initialization.

        Args:
            opt:
        """
        log = ROOT_ISL_DEVICE.get_child(__name__)
        changed = False

        log.debug("Checking value validity.")
        if not isinstance(opt, bool):
            raise ValueError()
        else:
            log.debug("Success")

        log.debug(f"Setting StoreAsFloat to {opt}")

        if self.__store_as_float != opt:
            changed = True
            self.__store_as_float = not self.__store_as_float

        if changed:
            self.number = float(self.number) if self.__store_as_float else int(self.number)
        else:
            log.error("Numerical.store_as_float contains the same value as second valve.")

    def __count_noun(
            self,
            noun: str,
            count: Optional[Union[int, float]] = None,
            only_noun=False,
            skip_commify=False,
            capitalize=False,
            to_words=False,
            full_stop=False,
            period=False,
            round_num=None,
            as_int=False
    ):
        """

        An alias for inflect.engine().plural_noun

        Returns the appropriate string for annotating counts of a given noun.

        Parameters:
            full_stop (bool):
                 Include a full-stop punctuation mark at the end of the final statement. Users that
                 are from America may know this punctuation mark better as 'period' (or; '.'. See
                 below for the 'period' parameter alias)
                 (Defaults to bool(False)

            period (bool):
                Include a period mark at the end of the final statement. Users from some areas may
                know this mark better as 'full-stop'

            noun (str):
                Returns a string with the appropriate statement iy,.hy,i for annotating counts of various nouns.

            count (int/float, optional):
                The number of 'noun' that exists.
                (Defaults to 'Numerical.number')

            only_noun (bool, optional):
                Return only the properly pluralized noun.
                (Defaults to False)

            skip_commify (bool, optional):
                (Ignored if either 'only_noun' or 'count_to_words' are True) Return the number as
                part of the statement with commas added to the proper places.
                (Defaults to False)

            to_words (bool, optional):
                (Ignored if 'only_noun' is True) Modify the first part of the statement
                (which contains the count) to words instead of numbers.
                (Defaults to False.)

            capitalize (bool, optional):
                Return the words for the count with the first letter capitalized.
                (Defaults to False, ignored if 'count_to_words' is False.)

            round_num (int, optional):
                Round the number returned to a given precision in decimal digits.
                (Defaults to None(NoneType), ignored if 'only_noun' is True.)
                
            as_int (bool, optional):
                Should number returned be an integer?
                (Defaults to bool(False), ignored if 'to_words' is True)

        Raises:

            ValueError:
                Raised when a value is provided for 'noun' that is not a string.

        """
        # Make sure our noun is a string or raise ValueError
        log = ROOT_ISL_DEVICE.get_child(__name__)
        if not isinstance(noun, str):
            raise ValueError("Noun must be of type: str")

        # Make sure 'count' is not 'None' or we can't do anything with it.
        #
        # If 'count' is indeed 'None' we'll use the number that seeded this instance of the
        # 'Numerical' class.
        if count is None:
            count = self.number
        if as_int:
            count = int(count)

        # If we received the parameter to round the results, we'll do that.
        if round_num is not None and isinstance(round_num, int):
            count = round(count, round_num)

        # Pluralize the noun string
        n_noun = INF.plural_noun(noun, count)

        # If the parameter 'to_words' is true, we need to convert the number to words (using the
        # 'to_words' function of this class before we finally concatenate our results.
        if to_words:
            c_count = self.to_words(count)
        else:

            # If we did not receive a bool(True) value for the 'skip_commify' parameter we'll
            # send our number off to be commified by self.commify before concatenation
            c_count = count if skip_commify else self.commify(target_number=count)
        # If the parameter 'only_noun' evaluates to bool(True) the request should not contain
        # the root number in any form.
        statement = n_noun if only_noun else f"{c_count} {n_noun}"
        # If the 'capitalize' parameter evaluates to bool(True) we capitalize the final statement, as
        # one capitalizes a sentence.8h
        if capitalize:
            statement = statement.capitalize()

        # If the parameter 'full_stop' evaluates to bool(True) a full-stop (or; 'period' to most of my
        # fellow Americans.
        if full_stop or period:
            statement += "."

        # Return the appropriate pluralized (or not) noun based off the number provided
        return statement

    def to_str(self):
        """

        Return 'self.number' as a string type without modifying it at  all from its original form.

        Returns:
            (int/float): The contents found at 'self.number' but in as a string type.

        """
        return str(self.number)

    def commify(self, target_number: int = None):
        """

        Return a string containing your number as a string with commas added

        Parameters:
            target_number (int, optional): The number you'd like returned as a commified string. Defaults to
            'Numerical.number'

        Returns:
            str: self.number converted to a string and with appropriately placed commas.

        """
        num = self.number if target_number is None else target_number
        return "{:,}".format(num)

    def to_words(self, target_num=None):
        """

        An alias to inflect.engine().number_to_words()

        Literally just passes 'target_num' to INF.number_to_words and returns the result of that

        Parameters:
            target_num (int, optional): The number you'd like returned in word-form. Defaults to `Numerical.number`.

        Returns:
            string: A string representing your target number in words.

        Raises:
            NotANumberError: Raised if 'target_num' isn't an integer or a float.

        """

        if target_num is None:
            target_num = self.number

        if isinstance(target_num, (int, float)):
            return INF.number_to_words(str(target_num))
        else:
            raise ValueError(
                    "The parameter 'target_num' needs to be an integer or a float"
            )

    def __radd__(self, other: (int, float), force_return_count=False, force_return_self=False,
                 replace_number=False):

        self.__check_other__(other)

        ans = other + self.number

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __add__(self, other: (int, float), force_return_count=False, force_return_self=False, replace_number=True):
        """

        Return self + other_num.

        Arguments:
            other (int|float):
                The number that you wish to add to `num` .

            force_return_count (bool, Optional):
                Return 'count_noun' after getting the sum.

                Optional, defaults to bool(False).

            force_return_self:
                Return the object.

                Optional, defaults to bool(False).

        .. warning:: Passing a value of bool(True) (or the evaluated equivalent) to both 'force_return_self' and
            'force_return_count' will result in a 'ValueError' exception being raised.

        Returns:
            float or int

        """
        self.__check_other__(other)

        ans = self.number + other

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __mul__(self, other, force_return_count=False, force_return_self=False, replace_number=True):

        ans = other if other in [0, 0.0] else self.number * other

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    @staticmethod
    def __check_other__(other):
        """

        Determines if the passed object can be treated like either a float or an integer. If so; returns NoneType,
        otherwise a 'HumanizeErrors.NumericalErrors.NotANumberError' exception is raised.

        Args:
            other:
                The object you'd like to check for compatibility.

        Returns:
            None

        Raises:
            HumanizeErrors.NumericalErrors.NotANumberError:
                Raised when a non-number object is passed as the value for the 'other' parameter.

        """
        try:
            int(other)
            float(other)
        except ValueError as e:
            raise NUM_ERR.NotANumberError(other) from e

    def __rmul__(self, other, force_return_count=False, force_return_self=False, replace_number=False):
        ans = None

        self.__check_other__(other)

        ans = other if other in [0, 0.0] else other * self.number

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __rtruediv__(self, other, force_return_count=False, force_return_self=False, replace_number=False):

        self.__check_other__(other)

        if other == 0 or self.number == 0:
            raise ZeroDivisionError()

        ans = other / self.number

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __truediv__(self, other, force_return_count=False, force_return_self=False, replace_number=True):

        self.__check_other__(other)

        if other == 0 or self.number == 0:
            raise ZeroDivisionError()

        ans = self.number / other

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __rsub__(self, other, force_return_count=False, force_return_self=False, replace_number=False):

        self.__check_other__(other)

        ans = other - self.number

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __sub__(self, other, force_return_count=False, force_return_self=False, replace_number=True):

        self.__check_other__(other)

        ans = self.number - other

        if replace_number:
            self.number = ans

        if force_return_count and self.__noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return ans

    def __sqrt__(self) -> Callable[[Optional[ContextManager]], Decimal]:
        return self.dec_num.sqrt

    def __str__(self):
        return self.to_words()

    def __float__(self):
        return float(self.number)

    def __int__(self):
        return int(self.number)

    def __repr__(self):
        return self.to_words()



from inspyre_toolbox.path_man import provision_path

PROFILE_CONFIGS = {
        'bash': '~/.bashrc',
        'zsh':  '~/.zshrc',
        'fish': '~/.config/fish/config.fish',
        }


def get_profile_path(profile: str = None) -> "pathlib.Path":
    """
    Get the path of a profile file.

    Parameters:
        profile (str):
            The profile to get the path of.

    Returns:
        str: The path of the profile file.
    """

    if profile is None:
        import os

        profile = os.environ.get('SHELL').split('/')[-1]

    path = provision_path(PROFILE_CONFIGS.get(profile))

    if path.exists():
        return path
    else:
        raise FileNotFoundError(f"Profile file not found: '{path}'")



import os

from inspyre_toolbox.sys_man.operating_system.linux import MOD_LOGGER as PARENT_LOGGER

MOD_LOGGER = PARENT_LOGGER.get_child('environment')

DISPLAY = 'DISPLAY' in os.environ


def has_gui_access():
    """
    Check if the current user has access to the GUI. This is so you can see if a GUI prompt can be displayed.

    Returns:
        bool:
            True if the user has access to the GUI, False otherwise.

    """
    # Check if the $DISPLAY environment variable is set
    if not DISPLAY:
        return False

    # Try to connect to the X server
    try:
        from Xlib import display as x_display

        display = x_display.Display()
        display.close()
        return True
    except Exception:
        return False


GUI_ACCESS = has_gui_access()

__all__ = [
        'GUI_ACCESS'
        ]



"""
Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/chrono/sleep.py

"""
import time

from inspyre_toolbox.syntactic_sweets.classes.decorators.type_validation import validate_type


class NegateSwitch:
    def __init__(self, initial_state=True):
        self.__state = None

    @property
    def state(self):
        return self.__state

    @state.setter
    @validate_type(bool)
    def state(self, new):
        self.__state = new




def sleep(seconds, precision=0.001, negate_obj=None):
    """
    A sleep function that can be interrupted by KeyboardInterrupt.

    Parameters:

        seconds (float):
            Number of seconds to sleep.

        precision (float):
            The interval in seconds to check for interruption. Default is 0.1 seconds.

        negate_obj (object, optional):
            `sleep` will change the value of this object to `False` if the sleep is interrupted. Default is None.

    Returns:
        None

    Raises:
        KeyboardInterrupt:
            If the user interrupts the sleep.

    Since:
        v1.6.0

    Example Usage:
        >>> from inspyre_toolbox.chrono import sleep
        >>> sleep(5)
        Sleeping for 5 seconds...
        (User interrupts after 3 seconds)
        Sleep interrupted by user.
    """
    start_time = time.time()
    print(f"[DEBUG] Starting interruptible_sleep for {seconds} seconds with precision {precision} seconds.")

    switch = True if negate_obj is None else negate_obj

    if switch is None:
        switch = NegateSwitch()
        switch.state = True

    while switch.state:
        try:
            remaining = seconds - (time.time() - start_time)
            print(f"[DEBUG] Remaining time: {remaining:.2f} seconds.")
            if remaining <= 0:
                print("[DEBUG] Sleep duration completed.")
                break
            time.sleep(min(precision, remaining))
        except KeyboardInterrupt:
            print("\n[DEBUG] Sleep interrupted by user.")

            if negate_obj is not None:
                negate_obj = False
            raise



from abc import ABC, abstractmethod
from typing import Dict, Optional

from inspyre_toolbox.config_man.options.ledger import OptionLedger, OptionLedgerEntry


# Base Option Class
class Option(ABC):
    """
    Base class for all options.
    """
    ledger = None  # This will be initialized when the first option is added

    def __init__(self, name: str, description: str, value: Optional[object] = None):
        self.name = name
        self.description = description
        self.value = value

    @abstractmethod
    def set_value(self, value):
        """Set the value of the option, enforcing type checking."""
        pass

    @abstractmethod
    def to_dict(self) -> Dict:
        """Convert the option to a dictionary."""
        pass

    @staticmethod
    @abstractmethod
    def from_dict(data: Dict):
        """Create an Option object from a dictionary."""
        pass

    def help(self):
        return self.description

    def log_change(self, old_value, new_value):
        """Log the change in the option's value to the ledger."""
        if Option.ledger is None:
            Option.ledger = OptionLedger()  # Create the ledger when the first option change is made
        entry = OptionLedgerEntry(self.name, old_value, new_value)
        Option.ledger.add_entry(entry)



import os

from inspyre_toolbox.log_engine.no_log import NoLog


def get_pid(name: str = None, inspy_logger_device=NoLog()) -> int:
    log = inspy_logger_device.get_child('get_pid')

    if not name:
        log.warning('No name provided, returning own PID')
        return get_own_pid()


def get_own_pid():
    return os.getpid()



"""
Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/type_man/string.py

Description:
    A custom string class that supports case-insensitive comparisons.
"""


class String(str):
    """
    A custom string class that supports case-insensitive comparisons.

    Parameters:
        value (str):
            The string value.

        compare_sensitively (bool, optional):
            If True, comparisons will be case-sensitive. Default is False.
    """
    def __new__(cls, value, *args, compare_sensitively=False, **kwargs):
        instance = super().__new__(cls, value)
        instance._compare_sensitively = compare_sensitively
        return instance

    @property
    def compare_sensitively(self):
        """
        Get the case sensitivity of the string.

        Returns:
            bool:
                True if the string is case-sensitive, False otherwise.
        """
        return self._compare_sensitively

    @compare_sensitively.setter
    def compare_sensitively(self, new):
        """
        Set the case sensitivity of the string.

        Parameters:
            new (bool):
                True if the string should be case-sensitive, False otherwise.:

        Returns:
            None
        """
        self._compare_sensitively = new

    def __eq__(self, other):
        me = str(self)
        if hasattr(other, 'compare_sensitively'):
            other = str(other)
        if not self.compare_sensitively:
            if hasattr(other, 'compare_sensitively'):
                other = str(other)
            me = me.lower()
            other = other.lower()
        return me == other



"""


Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/chrono/__init__.py
 

Description:
    

"""
from inspyre_toolbox.chrono.sleep import sleep


__all__ = ['sleep']



from collections import OrderedDict

from inspyre_toolbox.conversions.roman_numerals.errors import InvalidRomanNumeralStringError
from inspyre_toolbox.core_helpers.logging import ROOT_ISL_DEVICE, add_isl_child
from inspyre_toolbox.humanize import Numerical

ROOT_ISL_DEVICE.set_level('DEBUG')

LOG_NAME = 'conversions.roman_numerals'

MOD_LOG_DEVICE = add_isl_child(LOG_NAME)

LOG = MOD_LOG_DEVICE.logger

LOG.debug('Logger started.')

ROMAN_NUMERALS = OrderedDict(
        {
                'I':  1,
                'IV': 4,
                'V':  5,
                'IX': 9,
                'X':  10,
                'XL': 40,
                'L':  50,
                'XC': 90,
                'C':  100,
                'CD': 400,
                'D':  500,
                'CM': 900,
                'M':  1000
        }
)
"""
.. _rn_constant:

    A list of roman numerals that are constant.
    
    A list of roman numeral strings to compare against. The list is ordered by the 
    value of the roman numeral. The list is primarily used to determine if a provided string 
    is valid.
    
    
"""


def validate_roman_numeral_str(roman_numeral: str):
    """
    The 'validate_roman_numeral_str' function checks to see if the parameter 'roman_numeral' is a string.
    If it is, then each character in the string will be checked against 'ROMAN_NUMERALS'. If any characters are not
    found in :ref:`rn_constant`, an :class:`InvalidRomanNumeralStringError` exception will be raised

    Args:
        roman_numeral:str: Pass the roman numeral string to be validated

    Returns:
        True if the roman_numeral is a valid roman numeral string
    """
    log_device = ROOT_ISL_DEVICE.get_child(f'{LOG_NAME}.validate_roman_numeral_str')
    log = log_device.logger

    try:
        if not isinstance(roman_numeral, str):
            raise TypeError(f"The parameter 'roman_numeral' for 'validate_roman_numeral_str' must be of type 'str' "
                            f"not {type(roman_numeral)}'")

        for char in list(roman_numeral):
            if char not in ROMAN_NUMERALS:
                raise InvalidRomanNumeralStringError(roman_numeral, skip_print=True)
    except TypeError as e:
        log.error(e)
        return False
    except InvalidRomanNumeralStringError as e:
        log.error(e)
        return False
    return True


class RomanNumeral(object):

    def __init__(self, roman_numeral: str, noun: str = None):
        self.__roman_numeral = roman_numeral
        self.__integer = None
        self.__noun = noun
        self.__replace_noun = True
        self.__rn_changed = None

    @property
    def numeral_map(self):
        return ROMAN_NUMERALS

    @property
    def validate(self):
        return validate_roman_numeral_str(self.formatted)

    @property
    def provided(self):
        return self.__roman_numeral

    @provided.setter
    def provided(self, new_roman_numeral):
        nrn = new_roman_numeral.upper()
        if not validate_roman_numeral_str(nrn):
            raise InvalidRomanNumeralStringError(provided=nrn)
        self.__roman_numeral = nrn
        self.__rn_changed = True

    @property
    def formatted(self):
        return self.provided.upper()

    @property
    def noun(self):
        return self.__noun

    @noun.setter
    def noun(self, noun):
        if not isinstance(noun, str):
            raise ValueError(f"Provided property value for 'noun' must be of type 'str' not {type(noun)}")
        else:
            self.__noun = noun

    @property
    def as_numerical(self):
        return Numerical(self.as_int, noun=self.noun or None)

    @property
    def as_int(self):
        i = 0
        num = 0
        while i < len(self.formatted):
            if i + 1 < len(self.formatted) and self.formatted[i:i + 2] in ROMAN_NUMERALS:
                num += ROMAN_NUMERALS[self.formatted[i:i + 2]]
                i += 2
            else:
                num += ROMAN_NUMERALS[self.formatted[i]]
                i += 1
        self.__integer = num

        return self.__integer


def roman_numeral_to_integer(roman_numeral: str, noun: str = None, return_object=False, commify: bool = False):
    """
    The 'roman_numeral_to_integer' function converts a Roman Numeral to an integer.

    Args:
        roman_numeral: (str):
            Specify the roman numeral to be converted. (Required)

        noun: (str):
            Specify a noun for the roman numeral., (Optional, defaults to NoneType)

        return_object (bool):
            Determine whether the function should return an object or a string. (Optional; defaults to False)

        commify(bool)=False: Specify whether the numerical value should be returned with commas separating thousands

    Returns:
        The integer value of a roman numeral string.
    """
    if not isinstance(roman_numeral, str):
        raise TypeError(f'Parameter value for "roman_numeral" must be of type "str" not "{type(roman_numeral)}"')

    rn = RomanNumeral(roman_numeral, noun=noun)

    if return_object:
        return rn

    return rn.as_numerical.commify() if commify else rn.as_int



"""
This module contains functions to check the current operating system.

Functions:
    is_windows:
        Check if the current operating system is Windows.

    is_linux:
        Check if the current operating system is Linux.

Since:
    1.6.0
"""

import os


def is_windows() -> bool:
    """
    Check if the current operating system is Windows.

    Returns:
        bool:
            True if the current operating system is Windows, False otherwise.
    """
    return os.name == 'nt'


def is_linux() -> bool:
    """
    Check if the current operating system is Linux.

    Returns:
        bool:
            True if the current operating system is Linux, False otherwise.
    """
    return os.name == 'posix'


def is_macos() -> bool:
    """
    Check if the current operating system is macOS.

    Returns:
        bool:
            True if the current operating system is macOS, False otherwise.
    """
    return os.name == 'mac'


def is_supported() -> bool:
    """
    Check if the current operating system is supported.

    Returns:
        bool:
            True if the current operating system is supported, False otherwise.
    """
    return is_windows() or is_linux() or is_macos()



"""
This module contains functions that help determine the environment the program is running in.

Functions:
    has_gui_access -> bool:
        Check if the current user has access to the GUI.

Constants:
    GUI_ACCESS -> bool:
        True if the user has access to the GUI, False otherwise.

Exports:
    GUI_ACCESS -> bool:
        True if the user has access to the GUI, False otherwise.
"""
import ctypes
from ctypes import wintypes

from inspyre_toolbox.sys_man.operating_system.win32 import MOD_LOGGER as PARENT_LOGGER

MOD_LOGGER = PARENT_LOGGER.get_child('environment')


def has_gui_access():
    """
    Check if the current user has access to the GUI. This is so you can see if a GUI prompt can be displayed.

    Returns:
        bool:
            True if the user has access to the GUI, False otherwise.

    """
    user32 = ctypes.WinDLL('user32', use_last_error=True)
    if h_desktop := user32.OpenInputDesktop(0, False, wintypes.DWORD(0x0100)):
        user32.CloseDesktop(h_desktop)
        return True
    else:
        return False


GUI_ACCESS = has_gui_access()

__all__ = [
        'GUI_ACCESS'
        ]



"""


Author:
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File:
    inspyre_toolbox/decor/__init__.[y

Description:
    This module contains decorators that can be used to add functionality to classes and functions.

    The available decorators are:
        - method_alias: A decorator that allows you to add aliases to a class's method.
        - add_aliases: A decorator to add aliases to a class's methods.
        - validate_type: A decorator for validating the type and optionally the value of a value passed to a class
          property setter, with an option to convert to a preferred type if specified and to enforce value restrictions.
        - validate_func_args: A decorator for validating the type and optionally the value of arguments passed to a function.
"""
from functools import wraps
from typing import Optional

from .descriptors import RestrictedSetter


def convert_to_preferred_type(preferred_types):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            new_args = []
            for i, (arg, _) in enumerate(zip(args, preferred_types)):
                preferred_type = preferred_types[i]
                if preferred_type and not isinstance(arg, preferred_type):
                    try:
                        arg = preferred_type(arg)
                    except Exception as e:
                        raise TypeError(f"Could not convert argument {i} to preferred type {preferred_type.__name__}: {e}")
                new_args.append(arg)
            return func(*new_args, **kwargs)
        return wrapper
    return decorator


def validate_func_args(*validators):
    """
    A decorator for validating the type and optionally the value of arguments passed to a function.
    Each argument can have its own set of allowed types, a preferred type, and allowed values.

    Parameters:
        *validators (dict): Each dictionary in validators specifies the validation rules for one argument.
                            The keys in the dictionary can be:
                            - 'allowed_types': A tuple of allowed types for the argument.
                            - 'preferred_type': The preferred type to which the argument should be converted if possible.
                            - 'allowed_values': A list of values that the argument is allowed to have.
                            - 'case_sensitive': Specifies whether string comparisons should be case sensitive.

    Returns:
        A decorator function for the function.

    Raises:
        TypeError: If an argument does not match one of the allowed types or cannot be converted to the preferred type.
        ValueError: If the argument is of the correct type but not in the allowed values list.

    Example:
        >>> @validate_func_args(
        ...     {'allowed_types': (str,), 'allowed_values': ["hello", "world"], 'case_sensitive': False},
        ...     {'allowed_types': (int,), 'preferred_type': float}
        ... )
        ... def my_function(word, number):
        ...     print(f"{word} {number}")
        ...
        >>> my_function("HELLO", 123)  # This is valid and prints 'hello 123.0'
        >>> my_function("unknown", 123)  # This raises ValueError for the first argument
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            new_args = []
            for i, (arg, validator) in enumerate(zip(args, validators)):
                allowed_types = validator.get('allowed_types', ())
                preferred_type = validator.get('preferred_type')
                allowed_values = validator.get('allowed_values')
                case_sensitive = validator.get('case_sensitive', True)

                validate_type_decorator = validate_type(allowed_types)
                convert_type_decorator = convert_to_preferred_type(preferred_type)
                validate_value_decorator = validate_value(allowed_values, case_sensitive)

                arg = validate_type_decorator(lambda x: x)(arg)
                arg = convert_type_decorator(lambda x: x)(arg)
                arg = validate_value_decorator(lambda x: x)(arg)

                new_args.append(arg)

            return func(*new_args, **kwargs)

        return wrapper

    return decorator



def validate_value(allowed_values, case_sensitive=True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for i, (arg, _) in enumerate(zip(args, allowed_values)):
                if allowed_values[i] is not None:
                    if isinstance(arg, str) and not case_sensitive:
                        validation_passed = any(val.lower() == arg.lower() for val in allowed_values[i])
                    else:
                        validation_passed = arg in allowed_values[i]

                    if not validation_passed:
                        allowed_values_str = ', '.join(map(str, allowed_values[i]))
                        raise ValueError(f"Argument {i} must be one of {allowed_values_str}, got {arg}")
            return func(*args, **kwargs)
        return wrapper
    return decorator



class FrozenProperty(RestrictedSetter):

    def __init__(self, name, allowed_types=None, **kwargs):
        """
        Initialize the FrozenProperty object.

        Args:
            name (str): Name of the property.
            allowed_types (type or tuple of types, optional): Allowed types for the property value.

        Raises:
            AttributeError: If the initial value is not of the correct type.
        """
        super().__init__(name, allowed_types=allowed_types, **kwargs)

    def __call__(self, setter):
        def wrapped_setter(instance, value):
            if not instance.__dict__.get(f"_{self.name}_frozen", False):
                if self.allowed_types and not isinstance(value, self.allowed_types):
                    raise AttributeError(
                            f"Value must be of type {self.allowed_types}, not {type(value)}"
                            )
                setter(instance, value)
                instance.__dict__[f"_{self.name}_frozen"] = True
            else:
                raise AttributeError(f"Cannot modify a frozen property: {self.name}")

        return wrapped_setter


def frozen_property(name, allowed_types=None, restrict_setter=True):
    return FrozenProperty(name, allowed_types=allowed_types, restrict_setter=restrict_setter)



"""
This module contains decorators for restricting the change of classes.

Functions:

    restricted_to_internal_use:
        A decorator for restricting the use of a method to internal use only. This decorator raises a PermissionError if
        the function is called from outside the class, with an optional suggestion for an alternative method to use.

    restrict_change_to_default:
        A decorator for restricting the change of a property to only when it is set to the default value. This is kinda
        like a read-only property, but it allows changes only when the property is set to a specific value.
"""
import inspect
from functools import wraps


def restricted_to_internal_use(alternative_method=None):
    """
    A decorator for restricting the use of a method to internal use only.

    Parameters:
        alternative_method (str, optional):
            The name of an alternate method to suggest if the restricted method is called from outside the class.

    Returns:
        A wrapper function that raises a PermissionError if the function is called from outside the class.

    Raises:
        PermissionError:
            If the function is called from outside the class, a PermissionError is raised with an optional suggestion for an alternative method.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # Get the current frame and the caller's frame
            current_frame = inspect.currentframe()
            caller_frame = current_frame.f_back

            # Get the caller's class
            caller_class = caller_frame.f_locals.get('self', None).__class__

            # Check if the caller's class is the same as the class of the instance
            if caller_class is not self.__class__:
                suggestion = ""
                if alternative_method and hasattr(self, alternative_method):
                    suggestion = f" You may want to use the '{alternative_method}' method instead."
                raise PermissionError(f"The method {func.__name__} is restricted to internal use only.{suggestion}")

            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def restrict_change_to_default(default_value, allowed_types=None, allowed_values=None, change_error=None,
                               type_error=None, value_error=None, allow_deletion=False):
    """
    A decorator for restricting the change of a property to only when it is set to the default value and
    optionally checking for allowed types and allowed values. Optionally allows deletion of the property
    which resets it to the default value.

    Parameters:
        default_value (Any):
            The default value to which the property should be set in order to allow changes.
        allowed_types (tuple, optional):
            A tuple of allowed types for the property value.
        allowed_values (iterable, optional):
            An iterable of allowed values for the property.
        change_error (Exception, optional):
            The exception to raise if the property is not set to the default value.
        type_error (Exception, optional):
            The exception to raise if the property value is not of an allowed type.
        value_error (Exception, optional):
            The exception to raise if the property value is not an allowed value.
        allow_deletion (bool, optional):
            If True, allows deletion of the property which resets it to the default value.

    Returns:
        A decorator function for the property setter and deleter.

    Raises:
        change_error:
            If the property is not set to the default value, the specified error or AttributeError is raised.
        type_error:
            If the property value is not of an allowed type, the specified error or TypeError is raised.
        value_error:
            If the property value is not an allowed value, the specified error or ValueError is raised.
    """

    def decorator(setter):
        property_name = setter.__name__

        def wrapper(instance, value):
            if getattr(instance, property_name, default_value) != default_value:
                raise (change_error or AttributeError)(
                        f"This property cannot be changed unless it is set to the default value: {default_value}."
                        )

            if allowed_types is not None and not isinstance(value, allowed_types):
                raise (type_error or TypeError)(
                        f"The property value must be of type(s) {allowed_types}, but got type {type(value)}."
                        )

            if allowed_values is not None and value not in allowed_values:
                raise (value_error or ValueError)(
                        f"The property value must be one of {allowed_values}, but got {value}."
                        )

            setter(instance, value)

        def deleter(instance):
            setattr(instance, property_name, default_value)

        return property(fget=lambda instance: getattr(instance, property_name, default_value),
                        fset=wrapper,
                        fdel=deleter if allow_deletion else None)

    return decorator



import pkg_resources


def load_plugins():
    plugins = {}
    for entry_point in pkg_resources.iter_entry_points('inspyre_toolbox.plugins'):
        plugin = entry_point.load()
        plugins[entry_point.name] = plugin
    return plugins


def main():
    plugins = load_plugins()
    print("Available plugins:", plugins)



import os
import sys

from inspyre_toolbox.sys_man.operating_system.linux.users import is_admin


def run_as_admin():
    """
    Run the current script as an admin.

    Raises:
        PermissionError:
            If the current user is not an admin.

    Since:
        v1.6.0

    Example:
        >>> from inspyre_toolbox.sys_man.operating_system.linux.system import run_as_admin
        >>> run_as_admin()
    """
    if is_admin():
        raise PermissionError('The current user is already an admin.')

    cmd = ['sudo', sys.executable] + sys.argv
    os.execvp('sudo', ['sudo'] + cmd)


__all__ = [
    'run_as_admin'
]



#  Copyright (c) 2024. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech


import inspect
from typing import Any, Union


def is_class(obj: Any) -> bool:
    """
    Determine if an object is a class.

    Parameters:
        obj (Any):
            The object to check.

    Returns:
        bool:
            True if the object (:param:`obj`) is a class; False otherwise.

    Since:
        1.6.0

    Example:
        >>> is_class(str)
        True
        >>> is_class('string')
        False
    """
    return inspect.isclass(obj)


def is_instance(obj: Any) -> bool:
    """
    Determine if an object is an instance of a class, excluding classes themselves.

    Parameters:
        obj (Any):
            The object to check.

    Returns:
        bool:
            :bool:`True` if the provided object (:param:`obj`) is an instance of a class; :bool:`False`,
            otherwise.

    Since:
        1.6.0

    Example:
        >>> is_instance(str)
        False
        >>> is_instance('string')
        True
    """
    return not is_class(obj) and not isinstance(obj, type)


def commify(number: Union[int, float], separator: str = ',') -> str:
    """
    Convert a number to a string with thousands separators.

    This function formats integers and floating-point numbers into a string where thousands are separated by the specified character.

    Parameters:
        number (Union[int, float]):
            The number to convert. Must be an integer or floating-point number.
            separator (str): The character to use as the thousands' separator. Defaults to ','.

    Returns:
        str:
            The number as a string with the specified thousands separators.

    Raises:
        ValueError:
            If the input is not a valid number (int or float), or if the separator is an empty string.

    Since:
        1.6.0

    Example:
        >>> commify(1000000)
        '1,000,000'

        >>> commify(1000000, separator=' ')
        '1 000 000'

        >>> commify(-1234567.89, separator='.')
        '-1.234.567.89'
    """
    if not isinstance(number, (int, float)):
        raise ValueError("Input must be an integer or float")
    if not separator:
        raise ValueError("Separator must be a non-empty string")

    # Create a formatted string with the default comma separator
    formatted = f'{number:,}'

    # Replace commas with the specified separator
    return formatted.replace(',', separator)



from pathlib import Path
from typing import Union

from inspyre_toolbox.path_man import provision_path


def add_to_path(
        path: Union[str, Path],
        persistent: bool = False
        ):
    """
    Add a path to the PATH environment variable.

    Parameters:
        path (Union[str, Path]):
            The path to add to the PATH environment variable.

        persistent (bool):
            Whether to make the change persistent.

        profile_file (Union[str, Path):
            The profile file to add the path to.

        do

    Returns:
        None
    """


def add_to_path_file(
        path: Union[str, Path],
        path_file: Union[str, Path],
        do_not_provision_filepaths: bool = False,
        prepend: bool = False

        ):
    """
    Add a path to a path file.

    Parameters:
        path (Union[str, Path]):
            The path to add to the path file.

        path_file (Union[str, Path]):
            The path file to add the path to.

        do_not_provision_filepaths (bool):
            Whether to provision the file paths.

    Returns:
        None
    """
    if not do_not_provision_filepaths:
        path_file = provision_path(path_file)
        path = provision_path(path)

    if not path_file.exists():
        create_path_file(path_file)

    with open(path_file, 'a') as file:
        if prepend:
            file.write(f'\nexport PATH={path}:$PATH\n')
        else:
            file.write(f'\nexport PATH=$PATH:{path}\n')


def add_path_file_to_profile_file(profile_file: Union[str, Path], path_file: Union[str, Path]):
    """
    Add a path file to a profile file.

    Parameters:
        profile_file (Union[str, Path]):
            The profile file to add the path file to.

        path_file (Union[str, Path]):
            The path file to add to the profile file.

    Returns:
        None
    """
    profile_file = provision_path(profile_file)
    path_file = provision_path(path_file)

    if check_profile_file_for_path_file_load(profile_file, path_file):
        return

    with open(profile_file, 'a') as file:
        file.write('# Adding path file to profile file.\n'
                   f'\nsource {path_file}\n')


def check_profile_file_for_path_file_load(
        profile_file: Union[str, Path],
        path_file: Union[str, Path],
        do_not_provision_filepaths: bool = False
        ):
    """
    Check a profile file for a path file load.

    Parameters:
        profile_file (Union[str, Path]):
            The profile file to check for the path file load.

        path_file (Union[str, Path]):
            The path file to check for in the profile file.

        do_not_provision_filepaths (bool):
            Whether to provision the file paths.

    Returns:
        bool:
            Whether the path file is loaded in the profile file
    """
    if not do_not_provision_filepaths:
        profile_file = provision_path(profile_file)
        path_file = provision_path(path_file)

    with open(profile_file, 'r') as file:
        lines = file.readlines()

    return f'source {path_file}\n' in lines


def create_path_file(path: Union[str, Path]):
    """
    Create a path file.

    Parameters:
        path (Union[str, Path]):
            The path to create the path file at.

    Returns:

    """
    path = provision_path(path)
    if not path.exists():
        path.touch(exist_ok=False)

    with open(path, 'w') as file:
        file.write('# This file is used to store paths for the PATH environment variable.\n')

    check_profile_file_for_path_file_load('~/.bashrc', path)


def remove_from_path(path: Union[str, Path], do_not_notify: bool = False):
    """
    Remove a path from the PATH environment variable.

    Parameters:
        path (Union[str, Path]):
            The path to remove from the PATH environment variable.

        do_not_notify (bool):
            Whether to notify the environment of the changes.

    Returns:
        None
    """
    pass


def remove_from_path_file(
        path: Union[str, Path],
        path_file: Union[str, Path],
        do_not_provision_filepaths: bool = False
        ):
    """
    Remove a path from a path file.

    Parameters:
        path (Union[str, Path]):
            The path to remove from the path file.

        path_file (Union[str, Path]):
            The path file to remove the path from.

        do_not_provision_filepaths (bool):
            Whether to provision the file paths.

    Returns:
        None
    """
    if not do_not_provision_filepaths:
        path_file = provision_path(path_file)
        path = provision_path(path)

    if not path_file.exists():
        raise FileNotFoundError(f"The path file {path_file} does not exist.")

    with open(path_file, 'r') as file:
        lines = file.readlines()

    with open(path_file, 'w') as file:
        for line in lines:
            if path not in line:
                file.write(line)


def list_path_directories():
    """
    List the directories in the PATH environment variable.

    Returns:
        list:
            The directories in the PATH environment variable.
    """
    pass



class NeedsProcessingError(Exception):
    """
    Raised when an operation is attempted on a file collection that hasn't yet been processed.

    Attributes:
        message (str):
            The error message.

    Note:
        This exception can usually be resolved by calling the `process_files` method on the file-collection object
        before attempting the operation.
    """

    def __init__(self, message: str = 'The file collection has not yet been processed!'):
        self.message = message
        super().__init__(self.message)



from typing import List

from PIL import Image

__all__ = [
        'PIL_FORMATS',
        'PIL_EXTENSIONS',
        'IMAGE_FORMATS',
        'VALID_IMAGE_EXTENSIONS',
        ]


def __get_PIL_formats() -> List[tuple[str, str]]:
    """
    Get all the image formats supported by the Python Imaging Library (PIL).

    Returns:
        List[tuple[str, str]]:
            A list of supported image formats.
    """
    formats = Image.registered_extensions()

    return [(ext.replace('.', ''), Image.MIME[fmt]) for ext, fmt in formats.items() if fmt in Image.MIME]


PIL_FORMATS = __get_PIL_formats()
"""
A list of tuples containing information on the formats supported by the Python Imaging Library (PIL).

Each tuple contains two elements:
    1. The file extension.
    2. The MIME type.
    
Example:
    [('bmp', 'image/bmp'), ('dib', 'image/bmp'), ('dcx', 'image/pcx'), ('eps', 'application/postscript'),
    ('ps', 'application/postscript'), ('gif', 'image/gif'), ('im', 'image/x-portable-bitmap'), ('jpg', 'image/jpeg'),
    ('jpe', 'image/jpeg'), ('jpeg', 'image/jpeg'), ('pcd', 'image/pcd'), ('pcx', 'image/pcx'), ('pbm', 'image/x-portable-bitmap'),
    ('pgm', 'image/x-portable-graymap'), ('png', 'image/png'), ('ppm', 'image/x-portable-pixmap'), ('psd', 'image/vnd.adobe.photoshop'),
    ('tif', 'image/tiff'), ('tiff', 'image/tiff'), ('xbm', 'image/x-xbitmap'), ('xpm', 'image/x-xpixmap')]

Note:
    Here's some important things to note:
        - The file extension is the key.
            - The file extension is a string.
            - The file extension is in the format 'ext' (note the absence of a '.').
        - The MIME type is the value.
            - The MIME type is a string.
            - The MIME type is in the format 'type/subtype'.
"""

del __get_PIL_formats  # Cleanup

PIL_EXTENSIONS = [pair[0] for pair in PIL_FORMATS]
"""
A list of the file extensions supported by the Python Imaging Library (PIL).

Example:
    [
        'bmp', 'dib', 'dcx', 'eps', 'ps', 'gif', 'im',
        'jpg', 'jpe', 'jpeg', 'pcd', 'pcx', 'pbm', 'pgm',
        'png', 'ppm', 'psd', 'tif', 'tiff', 'xbm', 'xpm'
    ]
"""

__temp = []

for pair in PIL_FORMATS:
    info = pair[1]
    type, desc = info.split('/')

    if type == 'image':
        __temp.append(pair)

IMAGE_FORMATS = __temp
"""
A list of tuples containing information on the image formats supported by the Python Imaging Library (PIL).
"""

del __temp  # Cleanup

VALID_IMAGE_EXTENSIONS = [pair[0] for pair in IMAGE_FORMATS]
"""
A list of the file extensions for valid image formats supported by the Python Imaging Library (PIL).
"""



import ctypes


def is_admin():
    """
    Check if the current user is an admin.

    Returns:
        bool:
            True if the current user is an admin, False otherwise.

    Since:
        v1.6.0

    Example:
        >>> from inspyre_toolbox.proc_man.windows.users import is_admin
        >>> is_admin()
        True
    """
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except Exception:
        return False



import win32api
import win32con
import win32security

# Constants for integrity levels
SECURITY_MANDATORY_UNTRUSTED_RID = 0x00000000
SECURITY_MANDATORY_LOW_RID = 0x00001000
SECURITY_MANDATORY_MEDIUM_RID = 0x00002000
SECURITY_MANDATORY_HIGH_RID = 0x00003000
SECURITY_MANDATORY_SYSTEM_RID = 0x00004000


def is_process_elevated(pid):
    try:
        # Open a handle to the process
        hProcess = win32api.OpenProcess(win32con.PROCESS_QUERY_INFORMATION, False, pid)
    except Exception as e:
        print(f"Failed to open process {pid}: {e}")
        return False
    try:
        # Open the access token associated with the process
        hToken = win32security.OpenProcessToken(hProcess, win32con.TOKEN_QUERY)
    except Exception as e:
        print(f"Failed to open process token: {e}")
        win32api.CloseHandle(hProcess)
        return False
    try:
        # Retrieve the integrity level of the process
        label = win32security.GetTokenInformation(hToken, win32security.TokenIntegrityLevel)
        sid = label[0]
        # Get the integrity level RID
        integrity_level = win32security.GetSidSubAuthority(sid, 0)
        # Check if the integrity level indicates elevated privileges
        is_elevated = integrity_level >= SECURITY_MANDATORY_HIGH_RID
    except Exception as e:
        print(f"Failed to get token integrity level: {e}")
        is_elevated = False
    finally:
        # Clean up handles
        win32api.CloseHandle(hToken)
        win32api.CloseHandle(hProcess)
    return is_elevated



import os
from typing import List


def separate_path_str(path_str: str) -> List[str]:
    """
    Separate a PATH environment variable string into a list of directories.

    Parameters:
        path_str (str):

    Returns:
        List[str]:
            A list of directories in the PATH environment variable.
    """
    return path_str.split(os.pathsep)



class FreezePropertyError(Exception):
    """Exception raised when attempting to modify a frozen property."""
    pass


class FrozenProperty:

    def __init__(self, name, default=None):
        self.name = f"_{name}"
        self.default = default
        self.getter = None
        self.setter = None

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = getattr(instance, self.name, self.default)
        return self.getter(instance) if self.getter else value

    def __set__(self, instance, value):
        if getattr(instance, self.name, self.default) != self.default:
            raise FreezePropertyError(f"Cannot modify frozen property '{self.name[1:]}'")
        if self.setter:
            value = self.setter(instance, value)
        setattr(instance, self.name, value)

    def set_getter(self, func):
        self.getter = func
        return self

    def set_setter(self, func):
        self.setter = func
        return self


class FreezePropertyMeta(type):

    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        cls._frozen_properties = {}
        for key, value in dct.items():
            if isinstance(value, FrozenProperty):
                cls._frozen_properties[key] = value
        return cls

    def __setattr__(cls, key, value):
        if key in cls.__dict__.get('_frozen_properties', {}):
            raise FreezePropertyError(f"Cannot modify frozen property '{key}' at class level")
        super().__setattr__(key, value)



from .arguments import Arguments

PARSED_ARGS = Arguments().parsed


def main():
    return getattr(PARSED_ARGS, 'NEW PATH')

    # add_to_path(PARSED_ARGS.new_path, do_not_notify=PARSED_ARGS.do_not_notify, profile_file=PARSED_ARGS.file)



"""
This module contains functions that help determine the environment the program is running in.
"""
from inspyre_toolbox.sys_man import MOD_LOGGER as PARENT_LOGGER, is_windows

MOD_LOGGER = PARENT_LOGGER.get_child('environment')

if is_windows():
    pass



from inspyre_toolbox.common.about.package import PACKAGE_INFO
from inspyre_toolbox.common.about.pypi import PYPI_VERSION_INFO

AUTHORS   = PACKAGE_INFO['developer']['org']['developers']
PROG_NAME = PACKAGE_INFO['name']['full']
URLS = PACKAGE_INFO['urls']

from inspyre_toolbox.common.about.version import *

_VERSION = VERSION

VERSION   = _VERSION.parse_version()


__all__ = [
        'AUTHORS',
        'PACKAGE_INFO',
        'PROG_NAME',
        'PYPI_VERSION_INFO',
        'URLS',
        'VERSION'
]



"""

Contains errors for inspyre_toolbox.conversions.roman_numerals

"""


class InvalidRomanNumeralStringError(Exception):
    """
    .. _rn_constant:

    """
    message = "An invalid Roman numeral string was passed to the 'provided' parameter."

    def __init__(self, provided, message=message, skip_print=True):
        """

        Raised when a string is passed to 'RomanNumerals' that does not match one of the Roman numeral characters (or
        pair)

        Parameters:
            provided (str):
                The string provided to the 'RomanNumeral' class that's invalid.

            message (str):
                Any additional information the caller might want to be made clear in the exception being raised to the
                end-user. (Optional)

            skip_print (bool):
                Should the exception print it's 'message' contents. (Optional)

        """

        self.message = self.message

        if message:
            self.message += f"\nSome additional information from caller: {message}"

        if not skip_print:
            print(self.message)



"""
The software organization information for the project.
"""
from inspyre_toolbox.common.about.author.dev_team import DEVELOPER_TEAM


SOFTWARE_ORG = {
        'name': 'Inspyre-Softworks',
        'developers': DEVELOPER_TEAM,
        'urls': {
                'main':   'https://inspyre.tech',
                'github': 'https://github.com/Inspyre-Softworks',
                'gitlab': 'https://gitlab.com/inspyre-softworks',
                'pypi':   'https://pypi.org/user/tblackstone/',
                }
        }
""" The software organization information for the project. """


SOFTWARE_ORG_URLS = SOFTWARE_ORG['urls']
""" The software organization URLs for the project. """


SOFTWARE_ORG_NAME = SOFTWARE_ORG['name']
""" The software organization name for the project. """



"""
This module provides functions for managing users and groups on Linux systems.

Functions:
    - is_admin():
        Checks if the current user has administrative privileges.
"""
import os


def is_admin() -> bool:
    """
    Checks if the current user has administrative privileges.

    Returns:
        bool:
            True if the current user has administrative privileges, False otherwise.
    """
    return os.getuid() == 0




LEAD_DEVELOPER = {
        'name': 'Taylor-Jayde Blackstone',
        'email': '<t.blackstone@inspyre.tech>',
        'urls': {
                'github': 'https://github.com/tayjaybabee',
                'gitlab': 'https://gitlab.com/tayjaybabee',
        }
    }




DEVELOPER_TEAM = {
        'lead': LEAD_DEVELOPER,
        }


__all__ = [
    'DEVELOPER_TEAM',
    ]



"""
Project: Inspyre-Toolbox
Author: Inspyre Softworks - https://inspyre.tech
Created: 5/11/2023 @ 8:39 PM
File:
  Name: __init__.py
  Filepath: inspyre_toolbox/solve_kit
"""
from statistics import mean

from inspyre_toolbox.common.about.version import VERSION as __VERSION__
from inspyre_toolbox.conversions.bytes import ByteConverter
from inspyre_toolbox.log_engine import ROOT_LOGGER as PARENT_LOGGER

MOD_LOGGER = PARENT_LOGGER.get_child('solve_kit')

MOD_LOGGER.debug(f'Module loaded: solve_kit (inspyre_toolbox - v{__VERSION__})')

DOWNLOAD_SPEED_UNITS = {
        'bps':  {'unit': 'bit', 'factor': 1},
        'kbps': {'unit': 'kilobit', 'factor': pow(10, 3)},
        'mbps': {'unit': 'megabit', 'factor': pow(10, 6)},
        'gbps': {'unit': 'gigabit', 'factor': pow(10, 9)},
        'tbps': {'unit': 'terabit', 'factor': pow(10, 12)},
        'ebps': {'unit': 'exabit', 'factor': pow(10, 15)},
        'zbps': {'unit': 'zettabit', 'factor': pow(10, 18)},
        'ybps': {'unit': 'yottabit', 'factor': pow(10, 21)},
        }


def find_unit(unit: str, case_sensitive=False):
    """
    Find a unit in the DOWNLOAD_SPEED_UNITS dictionary by its unit name or abbreviation.

    Parameters:
        unit (str):
            The unit to find.

    Returns:
        dict:
            The unit data.
    """
    if not case_sensitive:
        unit = unit.lower()

    return next(
            (
                    v
                    for k, v in DOWNLOAD_SPEED_UNITS.items()
                    if k == unit or v['unit'] == unit
                    ),
            None,
            )


def how_many_until(current:list, target=.98, iter_limit=pow(10, 10), delay=0):
    global iters

    def _process_():
        global iters
        cur = list(current)
        iters = 0
        while mean(cur) < target and iters < iter_limit:
            cur.append(1)
            iters += 1
        if iters == iter_limit:
            print('Escaping seemingly infinitely recursive loop')
        return iters

    try:
        _process_()
    except KeyboardInterrupt:
        iter_limit = iters


def download_time(file_size_bytes: int, download_speed: int, download_speed_unit: str = 'bit', return_in_seconds=False):
    bc = ByteConverter(file_size_bytes, 'byte')
    speed_data = find_unit(download_speed_unit.lower())

    if not speed_data:
        raise ValueError(f"Invalid download speed unit: {download_speed_unit}")

    # Convert download speed to bits per second
    download_speed_bps = download_speed * speed_data['factor']

    # Calculate download time in seconds
    download_time_seconds = bc.bytes_to_bits() / download_speed_bps

    if return_in_seconds:
        return download_time_seconds

    # Convert download time to appropriate unit
    time_unit = 'second' if download_time_seconds < 60 else 'minute'
    time_value = download_time_seconds if time_unit == 'second' else download_time_seconds / 60

    return f"{time_value:.2f} {time_unit}"





"""
The MIT License (MIT)
Copyright © 2023 Inspyre Softworks - https://inspyre.tech
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""



from functools import wraps
from pathlib import Path

from .type_validation import validate_type


def validate_path(exists=False, create=False, callback=None, do_not_resolve=False, do_not_expanduser=False):
    """
    A decorator for validating a path.

    Parameters:
        exists (bool, optional):
            If True, ensures that the path exists.

        create (bool, optional):
            If True and the path does not exist, creates the path.

        callback (callable, optional):
            A callback function to call if the path does not exist. Ignored if 'create' is True.

        do_not_resolve (bool, optional):
               If True, does not resolve the path.

        do_not_expanduser (bool, optional):
                If True, does not expand the user in the path.

    Returns:
        A decorator function for the property setter.

    Raises:
        ValueError: If 'exists' is True and the path does not exist.
        FileNotFoundError: If 'exists' is True and 'create' is False, and the path does not exist.
        TypeError: If the value is not a Path object.

    Example:
        >>> class MyClass:
        ...     @property
        ...     def my_path(self):
        ...         return self._my_path
        ...
        ...     @my_path.setter
        ...     @validate_path(exists=True)
        ...     def my_path(self, path):
        ...         self._my_path = path
        ...
        >>> obj = MyClass()
        >>> obj.my_path = Path("/path/to/existing/folder")  # This is accepted and set
        >>> obj.my_path = "/nonexistent/folder"  # This raises TypeError
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, path):
            prop_name = f'{self.__class__.__name__}.{func.__name__}'
            if isinstance(path, str):
                path = Path(path)
            if not isinstance(path, Path):
                raise TypeError(f"`{prop_name}` must be a `pathlib.Path` or `string` object!")

            if not do_not_expanduser:
                path = path.expanduser()

            if not do_not_resolve:
                path = path.resolve()

            path_exists = path.exists()
            if exists and not path_exists:
                raise ValueError(f"{prop_name} must be an existing path, got {path}")

            if not path_exists and create:
                path.mkdir(parents=True, exist_ok=True)

            if not path_exists and callback and not create:
                callback(path)

            return func(self, path)

        return wrapper

    return decorator



from inspy_logger import InspyLogger

from inspyre_toolbox.common.about.package import PLATFORM_DIRS
from inspyre_toolbox.settings import log_level as INSPY_LOG_LEVEL

DEFAULT_LOG_DIR_PATH = PLATFORM_DIRS.user_log_path
DEFAULT_LOG_FILE_PATH = DEFAULT_LOG_DIR_PATH.joinpath('inspyre_toolbox.log')

ROOT_LOGGER = InspyLogger('InspyreToolbox', console_level=INSPY_LOG_LEVEL, file_name=DEFAULT_LOG_FILE_PATH)

ROOT_LOGGER.debug('Module loaded: log_engine')



from contextlib import contextmanager


@contextmanager
def flag_lock(instance, flag_name):
    """
    Context manager to set and unset a private flag attribute in an instance.

    This context manager ensures that a specified private flag attribute
    in the given instance is set to True when entering the context and
    reset to False upon exiting the context. This is useful for indicating
    that an operation is in progress within a specific scope.

    Parameters:
        instance (object):
            The instance containing the private flag attribute.

        flag_name (str):
            The name of the flag attribute (without leading underscore).

    Raises:
        AttributeError:
            If the instance does not have the specified private flag attribute.

    Example usage:
        handler = OperationHandler()
        with flag_lock(handler, 'flag'):
            # Perform some operation
            pass

    This will automatically set handler._flag to True within the context
    and reset it to False when exiting the context.
    """
    private_flag_name = f"_{flag_name}"
    if not hasattr(instance, private_flag_name):
        raise AttributeError(f"{instance} has no attribute '{private_flag_name}'")

    setattr(instance, private_flag_name, True)
    try:
        yield
    finally:
        setattr(instance, private_flag_name, False)



from inspyre_toolbox.syntactic_sweets.classes.decorators import validate_type, validate_path
from inspyre_toolbox.syntactic_sweets.classes.descriptors import RestrictedSetter

__all__ = [
    'validate_type',
    'validate_path',
    'RestrictedSetter'
]



"""
Command-line arguments for the ist-version-tool program.

This module defines a custom `ArgumentParser` class for parsing command-line arguments for the ist-version-tool program.

Attributes:
    PROG (str):
        The program name for the ArgumentParser.

    DESCRIPTION (str):
        The description for the ArgumentParser.

Classes:
    Arguments (class):
        A custom ArgumentParser class for the ist-version-tool program.

        This class inherits from `argparse.ArgumentParser` and customizes initialization to define specific command-line
        arguments needed for the version information tool. It provides a method to parse the command-line arguments and
        return the parsed namespace object.

        Properties:
            parsed (argparse.Namespace):
                Parses the command-line arguments if not already done and returns the parsed namespace. If the arguments
                have already been parsed, it returns the stored namespace object. Otherwise, it parses the arguments and
                stores the result in the `__parsed` attribute.

        Methods:
            __init__:
                Initializes the custom ArgumentParser with predefined program name and description, and sets up the
                command-line arguments.

            parsed:
                Parses the command-line arguments if not already done and returns the parsed namespace.

Example:
    To use the Arguments class in an external script, import it as follows:

    ```python
    from inspyre_toolbox.cli.ist_version_tool.arguments import Arguments
    ```

    You can then create an instance of the `Arguments` class to parse command-line arguments for the ist-version-tool program:

    ```python
    args = Arguments().parsed
    ```

    The `args` variable will contain the parsed command-line arguments as a namespace object.
"""

from argparse import ArgumentParser

# Constants for program name and description
PROG = 'ist-version-tool'
DESCRIPTION = 'Get information about the version and update availability of the inspyre-toolbox package.'

# Exported names
__all__ = ['Arguments']


class Arguments(ArgumentParser):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('prog', PROG)
        kwargs.setdefault('description', DESCRIPTION)
        super().__init__(*args, **kwargs)
        self.__parsed = None
        self.__build_subcommands()

    def __build_subcommands(self):
        subparsers = self.add_subparsers(
            dest='command',
            title='subcommands',
            metavar='<command>',
            parser_class=ArgumentParser
        )

        subparsers.add_parser('version', help='Show version information')
        update_parser = subparsers.add_parser('update', help='Check for updates')
        update_parser.add_argument('-c', '--check', action='store_true', help='Check for updates')
        update_parser.add_argument('-d', '--download', action='store_true', help='Download updates')

    @property
    def parsed(self):
        if self.__parsed is None:
            self.__parsed = self.parse_args()
        return self.__parsed



from pathlib import Path
from typing import Optional

from inspyre_toolbox.path_man import provision_path
from inspyre_toolbox.ver_man.classes import VersionParser as Version


def get_version_from_file(file_path: Path, do_not_provision=False, do_not_raise_errors=False) -> Optional[Version]:
    """
    Gets the version from a file.

    Parameters:
        file_path (Path):
            The file path from which to get the version.

        do_not_provision (bool, optional):
            If True, does not provision the path. Defaults to False.

        do_not_raise_errors (bool, optional):
            If True, does not raise an error if the file is not found. Defaults to False.

    Returns:
        Optional[Version]:
            The version.

    Raises:
        FileNotFoundError:
            If the file is not found and :param:`do_not_raise_errors` is :bool:`False`.

        Exception:
            If an error occurs while reading the file and :param:`do_not_raise_errors` is :bool:`False`.

    Since:
        v1.6.0

    Example Usage:
        >>> from inspyre_toolbox.ver_man.helpers import get_version_from_file
        >>> from pathlib import Path
        >>> file_path = Path('path/to/version/file')
        >>> version = get_version_from_file(file_path)
        >>> print(version)
        1.6.0
    """

    if not do_not_provision:
        file_path = provision_path(file_path)

    if not file_path.exists():
        if do_not_raise_errors:
            return None
        raise FileNotFoundError(f"File not found: {file_path}")

    try:
        with open(file_path, 'r') as f:
            return Version(f.read().strip())
    except Exception as e:

        # If the error should not be raised, return None
        if do_not_raise_errors:
            return None

        # Otherwise, raise the error with the original traceback.
        raise e from e  # Raise the exception with the original traceback


def get_version_string_from_file(file_path: Path) -> str:
    """
    Gets the version string from a file.

    Parameters:
        file_path (Path):
            The file path from which to get the version string.

    Returns:
    Since:
        v1.6.0
    """

    return str(get_version_from_file(file_path))


def read_version_file(file_path, do_not_provision=False):
    if not do_not_provision:
        file_path = provision_path(file_path)

    with open(file_path, 'r') as f:
        return f.read().strip().replace('\n', '')



"""


Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/type_man/comparators/string.py
 

Description:
    

"""


def case_insensitive_compare(string1: str, string2: str) -> bool:
    """
    Compare two strings in a case-insensitive manner.

    Parameters:

        string1 (str):
            The first string to compare.

        string2 (str):
            The second string to compare.

    Returns:
        bool: True if the strings are equal, False otherwise.

    Since:
        v1.6.0

    Example Usage:
        >>> from ins
